# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  

Created on Tue Aug  4 16:03:30 2020
    module will be deprecated soon !

-------
Classes :
    
    ReadFile : Extract informations of files with index 

    
        methods :
        ---------
            * read_file : Readile file for extractiong data 
                
            * assert_index_value : assert index for sclinf infrmations on file 

Function utils : 
    
    : ll2DMS: 
        fonction convert degree decimal to DD:MM:SS
    : roundVar : 
        round value with exactitude 
    *: convert_phase_from_pdSeries(df, columns=None)
        return dataframe 
    : ll_to_utm2 : 
        return utm_zone, utm_easting, utm_northing
    :convert_rotate_phase : 
        return exactitude float (phase value ) according to geometry principal phase value
    : convert_phase_from_pdSeries : 
        return  a specified pd.Series :df[columns]
"""
#==============================================================================
import os 

from pycsamt.utils import gis_tools as gis 
import numpy as np 
import pandas as pd 

from scipy.interpolate import interp1d

#==============================================================================


def ll2DMS(position_val): 
    """Conversion function : Convert degree decimals (long and lat) 
    value on string DD:MM:SS.
    
     params : 
    ------------
    **longitude or latitude value in degree decimals  ***
    
    returns : 
    ---------
    ***Value converted in  string format : DD:MM:SS
    """
    # verifie la condition si la valeur est bien 
    mm,ss="00","00"
    if position_val in [None, 'None']:
        return None
    if type(position_val) is str:
        raise ValueError("{0} not correct format, should be {{Degree decimals}}".format(position_val))
        return 
    else : 
        deg=int(position_val)
        decim=position_val- deg
        if decim==0 :
            return "{0}:{1}:{1}:".format(deg,mm,ss)
        elif decim !=0 : 
            minute=decim*60 
            dec_1=minute-int(minute)
            if int(minute)<10: 
                mm="0{}".format(int(minute))
            else : 
                mm="{}".format(int(minute))
            if dec_1==0 :
                return"{0}:{1}:{2}".format(deg,mm,ss)
            else : 
                sec=dec_1*60
                seconde=roundVar(sec)
                if int(seconde)<10: 
                    ss="0{}".format(int(seconde))
                elif int(seconde)==60:
                    min_=int(minute)+1
                    if min_<10:
                        return "{0}:0{1}:{2}".format(deg,min_,ss)
                    elif min_==60: 
                        deg=deg+1
                        mm="00"
                        return "{0}:{1}:{2}".format(deg,mm,ss)
                    else:
                        return "{0}:{1}:{2}".format(deg,min_,ss)
                else : 
                    ss="{}".format(int(seconde))
                return "{0}:{1}:{2}".format(deg,mm,ss)
    
def roundVar(value): 
    """
    Fonction rounds value with eatitude 
    Parameters : Float number 
    ----------
    Returns
    -------
    return value rounded.
    """     
    decim=value-int(value)
    if decim<=.5:
        val=int(value)
    elif decim>.5 : 
        new_value=value+.5
        val=int(new_value)
    return val

def ll_to_utm2( lat, lon,reference_ellipsoid=23):
    """
    Function redirect convertion lat, long to Easting , northing, 
    
    Parameters
    ----------
    lat : float
        latitude in degree decimals.
    lon : TYPE
        longitude in degree decimals 
    reference_ellipsoid : integer, Check module gis_tools:
         The default is 23.matching the "WGS-84"

    Returns
    -------
    utm_zone : string
        Zone of survey .
    utm_easting : float
        cordinates of the area survey in meters(UTM) .
    utm_northing : float
        cordinates of the of the area survey in meter (UTM) .

    """
    utm_zone, utm_easting, utm_northing = gis.ll_to_utm(reference_ellipsoid,lat,lon)
    
    return utm_zone, utm_easting, utm_northing

def convert_rotate_phase(value,rot_phase=180):
    """
    Parameters
    ----------
    value : float or int 
        DESCRIPTION.
    Returns
    -------
    value : float value 
        Geometry excatitude phase value.
    
    Examples :
    ---------
        >>> import numpy as np
         >>>np.random.seed(0)
        >>> ss=np.array([convert_rotate_phase(float(ii)) for\
                 ii in np.linspace(1,525,50) if float(ii)<=382] )
        ... print(ss,"\n")
        ... print(np.linspace(1,525,50))
    """
    rot_phase=int(rot_phase)

    assert type(value)==float or type(value)==int,\
        'Unacceptable value:"{}", must be a float or int.'.format(value)

    if value > 0:
        if value >=rot_phase:
          return float(value%rot_phase)  

    elif value < 0: 
        if abs(value) >=rot_phase :
          return float(value%(-rot_phase))  

    return float(value) 


def convert_phase_from_pdSeries(df, columns=None,rotate=180) :
    """
    Parameters
    ----------
    df: object 
        ***df ***           must be a pandas dataframe 
        ***columns ***.     A string : a specified columns of the dataframe 

    Raises
    ------
    ValueError
        Not a float number

    Returns
    -------
    df[columns]: PandaSeries
        returns a specified pd.Series.
    """

    if columns==None : 
        return None 
    
    try :
        df.reset_index(drop=True)
    except:
        raise TypeError('{} must be a pd.Series'.format(df[columns]))

    for index, val in enumerate (df[columns]):
        b=convert_rotate_phase(val,rot_phase=rotate)
        df[columns].iloc[index]=b    
    return df[columns]
        
class ReadFile:
    """
    Class to read simple file : 
        it uses file wich contains head 
        
    Attributes :
    ============
    
    *FilePath : str or bytes
                path of work directory 
        
    * file_to_read : str 
                    file object can be a *.csv, .txt , or .bln, .dat .
        
    * sep: NoneType
                    string for spliting data  : ex : "," or "." or ":" or ";"
                    defaut is None used conventional method object.split()
        
    * ex_index : int or tuple
                    extraction index for the data in the file (columns).
                    set tuple for many index extraction :
                        ex: ex_index=9 or ex_index = -1
                            ex_index=(9,2,0) or (-10,-2,3)
            
    * headline : boolean 
                    the head of the file . if the file contains no head , 
                    set to FALSE , if not default is TRUE.
                    
    * head_range : int
                    the range of the head in the file
        
     *slicing : tuple or tuple of list 
                    if headline =True , default slice is (1, None), it removes 
                    the first headline
                     sets option : 
                         slicing must be a list : 
                             slicing=([1,4,9],[2,7,None] )
                             first index range :[1,4,9] is the start 
                             second index range :[2,7,None] is the end 
                        ---> [1:2],[4:7],[9:]
        
    methods :
    =========
    
    * read_file : return a dictionary for the data extract : 
                    key is the ilist_{number of ex_index}:
                    value : list of ex_index  contains string elements. 
                    ex (ex_index=9): {ilist_09:['573.40', '566.80', '553.95',...,
                                    '424.35', '411.50', '423.75']}
                    ex(ex_index=(9,2,0)) : {index_09:[573.40', '566.80'...],
                                            index_02:['110.48647', '110.48690',...,
                                            index_00:['S00','S01','S03'',...]}
                                              
    * assert_index_value : return int value for indexing 
                    np.random.seed(0)
                    su=(None,None,None)
                    print(ff.assert_index_value(ta))
                    ch=[round(15*ii,3) for ii in np.random.rand(15)]
                    ch2=tuple(ch)
                    ss=ff.assert_index_value(ch2)
                    ts= [ff.assert_index_value(ii) for ii in su ]
                    print(ch,"\n",ss,'\n',ts)
            
    * search_file : return datafile 
                    if FilePath is None ,search the Path of matching file
                        on the current directory. 
                    if file_to_read is None , search file on current directory. 
            
         
    Test: 
        >>> os.chdir(os.getcwd())
        >>> ff=ReadFile(FilePath =None, 
                       file_to_read="K1_exp.bln",
                       head_range=None,
                       ex_index=(8,-1),
                       slicing=([1,4,9],[2,7,None]))
        >>>  dd=ff.read_file()
        >>>  d1=[float(ss) for ss in dd['index_08']]
        >>>  d2=[float(ss) for ss in dd['index_09']]
        >>>  d1,d2=np.array(d1),np.array(d2)
        >>>  ary=np.concatenate((d1.reshape((1,d1.shape[0])),\
                                d2.reshape((1,d2.shape[0]))),axis=0)
        >>>  print(ary) 
    """
    
    
    def __init__(self, **kwargs):
        
        self.path=kwargs.pop("FilePath",None)
        self.data_index=kwargs.pop("ex_index",None)
        self.separate=kwargs.pop("sep",None)
        self.headline=kwargs.pop("headline",True)
        self.range_head=kwargs.pop("head_range",None)
        self.slicing=kwargs.pop("slicing",(None,None))
        
        self._flag,self._comp,self._emp=-1,-1,[]
        self._option, self._start, self._end=-1,None,None
                  
        if self.path is not None :
            if os.path.exists(self.path):
                self.file=kwargs.pop("file_to_read",None)
                if self.file is not None :
                    assert os.path.isfile(self.file),"{} don't exist"\
                        " in this directory.Please change your directory".\
                            format(self.file)
                    self.file=self.file
                elif self.file is None :
                    self.file=self.search_file_(self.path)
            else :
                raise FileExistsError('No such file in such directory')
                
        elif self.path is None :
            self.file=self.file=kwargs.pop("file_to_read",None)
            if self.file is None :
                self.file=self.search_file_(self.path)
                
        self.data_index=np.array(self.data_index)
        
        if np.iterable(self.data_index):
            self.indexl=self.data_index.tolist()
            for i,value in enumerate (self.indexl) :
                if type(value)==str :
                    self.indexl[i]=int(value)
            self.max_index=abs(np.asarray(self.indexl)).max()    
            self.indexl=[int(ii) for ii in self.indexl \
                    if (type(ii)==int) or (type(ii)==float)]
            self._flag=1

        if np.iterable(self.data_index)==False :
            self.indexl=self.data_index
            if type(self.indexl) == float:
                self.indexl=int(self._index) 
            self._flag=0
            
        self._start=self.assert_index_value(self.slicing[0])
        self._end=self.assert_index_value(self.slicing[1])
        
        if self.headline == True:
            if self.range_head is None :
                self.range_head=1
            elif self.range_head is not None :
                assert type(self.range_head)==int or type(self.range_head)==float,\
                    "Unacceptable value {}, must be integer".format(self.range_head)
                self.range_head=self.assert_index_value(self.range_head)
                
        if np.iterable(self._start) or np.iterable(self._end):
            
            assert len(self._start)==len(self._end),"slicing argument {0},"\
                " start= {1} and end= {2} must be the same length".\
                    format(self.slicing, self._start, self._end)  
                    
            if type(self._start)==tuple or type(self._end)==tuple :
                self._start, self._end=list(self._start),list(self._end)
            self._option=1

        elif np.iterable(self._start)==False or np.iterable(self._end)==False :
            self._option=0
            
    @property
    def separate(self):
        """check property attribute"""
        return self._separate
    
    @separate.setter
    def separate(self,value):
        if type(value)==float or type(value)==int :
            raise TypeError('Wrong attribute value: {0}.'\
                            ' Must be a string. Separate file Options are :'\
                                '";" or "." or ":" or "."'.format(value))
        self._separate= value
        

    def __str__(self):
        print("Welcome To Readline Class:")
        ReadFile.__doc__

    def read_file(self):
        """
        Methode to extract data in the file from index 
        
        parameters :
        -----------
            * headline : Boolean 
                if headline =True , default slice is (1, None), 
                it removes the first headline
                 sets option slicing : 
                     slicing must be a list : 
                         slicing=([1,4,9],[2,7,None] )
                         fisrt index :[1,4,9] is the start 
                         second index :[2,7,None] is the end 
                    ---> [1:2],[4:7],[9:]
        Returns
        -------
        dictionary of index data
        
        Examples:
        --------
           >>> import os 
           >>> import numpy as np
           >>> path=os.getcwd()
           dir =os.path.basename(path)
           >>> dic_ex=read_file(dir,(0,-9,3,-2,6),
                            headline=True,sep=" ")
          >>> sta=dic_ex['index_00']
          >>> azim=np.array([float(ss) for ss in dic_ex['index_08']])
          >>> print(azim)
        """
        
        # if self.file ==None :
        #     raise FileExistsError("No file in this directory: {}".format(os.getcwd()))
        if self.indexl== None :
            raise IndexError("No index for extraction! Index must be int, float , or tuple")
        
        try:
            with open (self.file,'r') as f :
                listfiles=f.readlines()
        
        except :
            FileNotFoundError('No such file in this directory: {0}.'.format(os.getcwd()))
        
        if self.headline:
            listfiles=listfiles[self.range_head:]
        
        # Option of slicing
        if self._option==0 :

            if self._start==None and self._end !=None :
                listfiles=listfiles[:self._end]
            elif self._start !=None and self._end ==None :
                listfiles =listfiles[self._start:]
            elif self._start != None and self._end != None \
                and type(self.slicing) != list :
                    assert np.array(self._start).min() < \
                np.array(self._end).max()<= len(listfiles)-1," Index {0} is out of the range."\
                    " Maximum index is {1}".format(np.array(self._end).max(),len(listfiles)-1)

                    listfiles=listfiles[self._start:self._end]
                    
        if self._option ==1:

            #assert slicing  value
            max_temp_start,max_temp_end=[],[]
            for ii in list(self._start) :
                if ii==None :
                    continue 
                max_temp_start.append(ii)
            for ii in list(self._end):
                if ii==None :
                    continue 
                max_temp_end.append(ii)
            
            assert np.array(max_temp_start).min() < \
                np.array(max_temp_end).max()<= len(listfiles)-1," Index {0} is out of the range."\
                    " Maximum index is {1}".format(np.array(max_temp_end).max(),len(listfiles)-1)
               
            tempfiles=listfiles[self._start[0]:self._end[0]]  
            for ii,val in enumerate(self._end) :
                if ii==0:
                    pass
                if ii !=0 :
                    if val == None :
                        assert (ii==len(self._start)-1) or \
                            (ii==len(self._end)-1), "Wrong mode of slicing !'None'"\
                                " nust be the first index or the last index of slicing"
                        temp= listfiles[self._start[ii]:]

                        for ss in temp :
                            tempfiles.append(ss)
                        temp=[]
                    else :
                        temp=listfiles[self._start[ii]:self._end[ii]]
                        for ss in temp :
                            tempfiles.append(ss)
                        temp=[]

            listfiles=tempfiles
         
        for ii, val in enumerate(listfiles) :
            self._comp=self._comp+1         # Avoid dictionnary reinitialization 
            if self.separate==None: 
                line=val.split()
            else : 
                line=val.split(self.separate)
            head=["index_{:02}".format(ind) for ind in range( len(line))]
    
            # extract multiple value from index 
            if self._flag==1 :
                assert 0 <= self.max_index < len(head),\
                    "Argument[ex_index={0:}]Input index {1:^3} is out of the range."\
                        "|Maximum index :{2:^3}|".\
                        format(tuple(self.indexl), self.max_index,len(head)-1)
                
                if self._comp==0:
                    headtemp=[head[ss] for ss in self.indexl]
                    none=[[]for ii in range(len(headtemp))]
                    dico={headtemp:none for headtemp,none in zip (headtemp,none)}
                for idx in self.indexl:
                    for key in dico: 
                        if 'index_{:02}'.format(idx) in key:
                            dico[key].append(line[idx])
                    if idx < 0 and (abs(idx)<=len(head)):
                        reverse=head[idx]
                        for key in dico :
                            if reverse in key :
                                dico[key].append(line[idx])
                                
            # extract single value from  index               
            if self._flag==0:
                if self.indexl<0 :
                    assert 0> self.indexl >= -len(head),"Index {0:^4} out of the range."\
                        " Minimum index egal to {1:^3}.".format(self.data_index,-len(head))
                        
                    reverse=head[self.indexl]
                    if self._comp==0:
                        dico={head[self.indexl]:[]}
                    for key in dico:
                        if reverse in key :
                            dico[key].append(line[self.indexl])
                if self.indexl>=0:
                    assert 0<= self.indexl < len(head),"Index {0:^4} out of the range."\
                        " Maximum index egal to {1:^3}.".format(self.data_index,len(head)-1)
    
                    for idx, car in enumerate( line) :
                        if idx==self.indexl:
                            self._emp.append(car)
                            dico={head[idx]:self._emp}
        return dico
    
    def assert_index_value(self, value):
        """
        Method to assert index value
        
        parameter:
        ----------
            * value : int, list , tuple, float or string 

        Returns
        -------
        value of index 
        
        Example :
        --------
            ff=ReadFile(file_to_read="K1_exp.bln", 
            ex_index=(8,-1),
            slicing=((2,15),(10,None)))
            np.random.seed(0)
            su=[12,[2],(3)]
            ch=[round(15*ii,3) for ii in np.random.rand(15)]
            ch2=tuple(ch)
            ss=ff.assert_index_value(ch2)
            ts= [ff.assert_index_value(ii) for ii in su ]
            print(ch,"\n",ss,'\n',ts)
        """
        # f=-1
        temp=[]
        
        if np.iterable(value)==False:
            
            if value == None:
                value=None
            else :
                value=int(value)
                
        if np.iterable (value):
            typi=type(value)
    
            if None in value :
                f=0
            elif None not in value :
                f=1
                
            if f==0:
                value=list(value)
                for ii in value : 
                    if ii==None :
                        # pass 
                        temp.append(ii)
                    else :
                        temp.append(int(ii))
                    
                if typi==tuple:
                    
                    value=tuple(temp)
                else :
                    value=temp
                    
            elif f==1 :
                value=[int(ii) for ii in value]
                if typi==tuple :
                    value=tuple(value)
        return value
    
    def search_file_(self, Pathfile):
        """
        Method to seach the matching file. 
        
        Returns
        -------
        *select_file * : str or basestring
            Path is None:
                search the list of the file present in the current work directory.
            file is None :
                input the matching file :
                users must choose the data file he want to extract. 
        Ex :
        ---
            >>> ff=ReadFine()
            Path=ff.path
            >>> None 
            >>> ds=ff.search_file(Pathfile=Path)
            ... ----------------------------------------------------------
                             list of current files                              
                Current directory : F:\OneDrive\Python\
                    CodesExercices\ex_avgfiles\modules
                Number of files found :   20
                filenames :(...)
                Enter the data filename:K1_exp.bln
                -----------------End of searching file.-----------------
        """
        temp=[]
        if Pathfile is None :
            Pathfile=os.getcwd()
            
        listfiles=os.listdir(Pathfile)
        for ss in listfiles :
            if os.path.isfile(ss):
                temp.append(ss)
            else  :
                continue
            
        filenames =temp
        print("-"*80)
        print('{:^80}\n'.format("list of current files"))
        print("Current directory : {:<}".format(Pathfile))
        print("Number of files found : {:>4}".format(len(filenames)))
        print("filenames :")
        
        for ii in filenames :
            print("{:<30}".format(ii) ,end='') 
        print("-"*80)
        select_file=input("Enter the data filename:")
        print('{:-^80}\n'.format("End of searching file."))
        
        if select_file in filenames :
            return select_file
        else :
            raise FileNotFoundError("No such file in this directory.")
            
            
def interpol_Scipy (x_value, y_value,x_new, kind="linear",
                    fill="extrapolate"):
    
    f=interp1d(x_value, y_value, kind=kind,fill_value=fill)
    y_new=f(x_new)
    
    return y_new
                


 


# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  
.. _module-Interface::`pycsamt.geodrill.geoDB.interfaceDB`
 
    :synopsis: Specially dedicate to Manage SQL 
Created on Tue Oct 13 15:28:57 2020

@author: @Daniel03
"""
import os, shutil 

import warnings
import pandas as pd 
# from pg8000 import DBAPI

import sqlite3 as sq3 

#
try : 
    from pycsamt.utils  import exceptions as CSex
    from pycsamt.utils._csamtpylog import csamtpylog
    _logger=csamtpylog.get_csamtpy_logger(__name__)
except :
    pass


class ManageDB(object) : 
    """
    build a datable postgre Sql  from dict_app.py  simple way to make a transit
    between two objects One object dict_app to populate DataBase
        
    Arguments
    ------------
           **db_name** : str  
               name of dataBase 
           **db_host** : st 
                path to database 

    ====================  ==============  ==================================== 
    Attributes              Type            Explanation 
    ====================  ==============  ==================================== 
    connex                  object            DataBase connection 
    curs                    object            Database cursor
    ====================  ==============  ==================================== 

    ==========================  ===============================================
     Methods                     Explanation 
    ==========================  ===============================================
    dicT_sqlDB                  send infos as  dictionnary to dataBase 
    execute req                 execute a sql_request
    drop_TableDB                drop all Tables in sql memory DB or single Table 
    closeDB                     close after requests the connection and the cursor
    commit                      transfer the data to DataBase. if not the data 
                                will still in the cursor and  not in the dataBase 
    print_last_Query            print the last operating system 
    export_req                  export the request on datasheet like excelsheet . 
    ==========================  ===============================================
        
        
    :Example: 
        
        >>> from sqlrequests import SqlQ
        >>> path= os.getcwd()
        >>> nameofDB='memory.sq3'
        >>> manDB=ManageDB(db_name=nameofDB, 
        ...                   db_host=path)
        ... print(SqlQ.sql_req[-1])
        ... manDB.executeReq(SqlQ.sql_req[2])
        ... ss=manDB.print_last_Query()
        ... print(ss)
        ... manDB.export_req(SqlQ.sql_req[-1],
                             export_type='.csv')
        ... manDB.dicT_sqlDB(dictTables=Glob.dicoT, 
                           visualize_request=False)
        """
    def __init__(self, db_name =None, db_host=None): 
        self._logging=csamtpylog.get_csamtpy_logger(self.__class__.__name__)
        
        self.db_host=db_host 
        self.db_name=db_name
        
        if self.db_name is not None  : 
            self.connect_DB()
            
            
    def connect_DB(self, db_host=None , db_name=None): 
        """
        Create sqqlite Database 
        
        :param db_host:   DataBase location path
        :type db_host: str
        
        :param db_name: str , DataBase name 
        :type db_name: str 
        """
        if db_host is not None : self.db_host = db_host 
        if db_name is not None : self.db_name = db_name 
        
        mess= ''
        if self.db_name is None  :
            mess ='Could not create a DataBase ! Need to input the DataBase name.'
            
        if self.db_host is None : 
            mess ='Could not create a DataBase : No "{0}" Database path detected.'\
                ' Need to input  the path for Database location.'.format(self.db_name)
        
        if mess !='': 
            warnings.warn(mess)
            self._logging.error(mess)
            # raise CSex.pyCSAMTError_SQL(mess )
         
        # try to connect to de dataBase 
        if self.db_host is not None :
            try : 
                self.connexDB=sq3.connect(os.path.join(self.db_host, self.db_name))
            except :
                warnings.warn("Connection to SQL %s failed !." %self.db_name)
                # raise CSex.pyCSAMTError_SQL_manager('Connection to SQL dataBase failed ,Please try again.') 
                
                self.success=0
            else : 
                self.curs =self.connexDB.cursor()
                self.success=1
        
    def dicT_sqlDB(self, dicTables, **kwargs): 
        """
        Method to create Table for sqlDataBase . 
        Enter Data to DataBase from dictionnary. 
        Interface objet : Database _Dictionnary  
        to see how dicTable is arranged , may consult dict_app module 
                
        Parameters
        ----------
        * dictTables : dict
               Rely on dict_app.py module. it populates the  datababse 
               from dictionnay app 
               
        Returns
        ---------
          str 
            execute queries from dict_app 
            
        :Example :
            
            >>>  mDB=GestionDB (dbname='memory.sq3, 
            ...                   db_host =os.getcwd()')
            >>> mDB.dicT_sqlDB(dicTables=Glob.dicoT,
            ...                  visualize_request=False)
            >>> ss=mB.print_last_query()
            >>> print(ss)
        """
        visual_req=kwargs.pop('visualize_request', False)
        
        field_dico ={'i':'INTEGER',"t":"TEXT",'b':'BOOL',
                     'd': 'HSTORE',"k": "SERIAL", 'n':'NULL',
                     "f": 'REAL','s':'VARCHAR','date':'date',
                     'by':'BYTHEA','l':'ARRAY',
                     }

        for table in dicTables: 
            req="CREATE TABLE %s (" % table
            pk=""
            for ii, descrip in enumerate(dicTables[table]): 
                field=descrip[0]
                tfield=descrip[1] # Type of field 
                # for keys in  field_dico.keys():
                if tfield in field_dico.keys():
                    # if tfield == keys : 
                    typefield=field_dico[tfield]
                else :
                    # sql vriable nom :'s':'VARCHAR'
                    typefield='VARCHAR(%s)'%tfield
                        
                req= req+'%s %s, ' %(field, typefield)
                

            if pk=='': 
                req=req [:-2] + ")" # delete the last ',' on req.
            else : 
                req =req +"CONSTRAINT %s_pk PRIMARY KEYS(%s))" %(pk,pk)
                
            if visual_req is True : # print the request built .
                print(req)                
            try : 
                self.executeReq(req)
            except : 
                pass # the case where the table already exists. 
            
        return req

    
    def executeReq(self, query, param=None):
        """
        Execute request  of dataBase  with detection of error. 
        
        Parameters  
        -----------
        * query : str  
                sql_query 
        * param : str 
                Default is None . 
        
        raise : 
        -------
            layout the wrong sql queries . 
            
        return : 
        -------
         int  
             1, the request has been successuful run .
              
        :Example: 
            
                >>>  for keys in Glob.dicoT.keys(): 
                ...        reqst='select * from %s'%keys
                >>>  ss=manageDB.executeReq(query=reqst)
                >>>  print(ss)
        """
        
        try :
            if param is None :
                self.curs.execute(query)# param)
            else :
                self.curs.execute(query, param)
                
        except: 
            warnings.warn(f'Request SQL {query} failed. May trouble of SQL server connexion. '\
                          'Please try again later ')
            # raise (f'Request SQL {query}executed failed',err)
            return 0
        else : 
            return 1
    
    def drop_TableDB(self, dicTables, drop_table_name=None ,
                     drop_all=False): 
        """
        Drop the name of table on dataBase or all databases.

        Parameters
        ----------
        * dicTables : dict
                application dictionnary. Normally provide from 
                dict_app.py module 
            
        * drop_table_name : str, optional
                field name of dictionnay (Table Name). 
                The default is None.
            
        * drop_all : Bool, optional
                Must select if you need to drop all table. 
                The default is False.

        Raises
        ------
            Exception : Errors occurs ! .

        """
        
        if drop_all is False and drop_table_name is None : 
            raise 'Must be input at least one name contained of keys in the dicT_app'
        elif drop_all is True : 
            for keys in dicTables.keys():
                req="DROP TABLE %s" %keys
                self.executeReq(req)
        elif drop_table_name is not None : 
            if drop_table_name in dicTables.keys(): 
                req="DROP TABLE %s" % drop_table_name
            else : 
                raise'No such name in the dictionnary application Table!'\
                    'Dict_app keys Tables Names are : ¨{0}'.format(dicTables.keys())
                    
            self.executeReq(req)
            
        self.connexDB.commit()
    
    def closeDB(self):
        """
        simple method to close Database. 
        """
        
        if self.connexDB : 
            # self.curs.close()
            self.connexDB.close()
            
    def commit(self) :
        """
        special commit method for the database when cursor  and connexion 
        are still open.
        """
        if self.connexDB : 
            self.connexDB.commit()
    
    def print_query(self, column_name=None ) : 
        """
        return the result of the previous query.
        
        Parameters :
        ------------
            * query_table_nam : str
                name of table to fetch colounm data .
        """
        if  column_name is not None :
            return self.curs.fetchone()
        else :
            return self.curs.fetchall()
    
    def export_req(self, query =None ,
                   export_type='.csv', **kwargs):
        """
        method to export data  from DataBase 

        Parameters
        ----------
        * query : str, optional
            Sql requests. You may consult sql_request files. 
            The default is None.
        * export_type : Str, optional
            file extension. if None , it will export on simple file. 
            The default is '.csv'.
        * kwargs : str
            Others parameters.

        Raises
        ------
        Exception
            Print wrong sqlrequests.

        :Example:
            
            >>> from sqlrequests import SqlQ
            >>> manageDB.executeReq(SqlQ.sql_req[2])
            >>> ss=manageDB.print_last_Query()
            >>> print(ss)
            >>> manageDB.export_req(SqlQ.sql_req[-1],
                                export_type='.csv',
                                )
        """
        
        exportfolder=kwargs.pop('savefolder','savefiles')
        filename =kwargs.pop('filename','req_file')
        indexfile=kwargs.pop('index',False)
        headerfile=kwargs.pop("header",True)
        
        if query is None : 
            raise Exception ("SQL requests (%s) no found ! Please Try to put "\
                             " your sql_requests"% query)
        elif query is not None : 
            
            df_sql=pd.read_sql_query(query,self.connexDB)
            
        if filename.endswith('.csv'):
            export_type ='.csv'
            filename=filename[:-4]
        elif filename.endswith(('.xlxm', '.xlsx', '.xlm')): 
            export_type='.xlsx'
            filename=filename[:-5]
        else : 
              assert export_type  is not None , 'Must input the type to export file.'\
                  ' it maybe ".csv" or ".xlsx"'

        #-----export to excel sheet  
        if export_type in ['csv','.csv', 'comma delimited',
                             'comma-separated-value','comma sperated value',
                                       'comsepval']:
            # export to excelsheet:  
            df_sql.to_csv(filename+'.csv', header=headerfile,
                  index =indexfile,sep=',', encoding='utf8')
            
            sql_write =1 
        elif export_type in ['xlsx','.xlsx', 'excell',
                             'Excell','excel','Excel','*.xlsx']: 
            df_sql.to_excel(filename+'.xlsx',sheet_name=filename[:-3],
                index =indexfile)
            sql_write =0

        #wite a new folder
        if exportfolder is not None : 
            try : 
                os.mkdir('{0}'.format(exportfolder))    
            except OSError as e :
                print(os.strerror(e.errno))
        sql_path=os.getcwd()
        savepath=sql_path+'/{0}'.format(exportfolder)
        
        if sql_write ==1 : 
            shutil.move(filename +'.csv', savepath)
            print('---> outputDB_file <{0}.csv> has been written.'.format(filename))
        elif sql_write ==0 :  
            shutil.move(filename +'.xlsx',savepath)
            print('---> outputDB_file <{0}.xlsx> has been written.'.format(filename))
        
        


    
    
    
                  
                        
                    
            
        



 



# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  
.. _module-Infos::`csamtpy.etc.infos`

    :synopsis: Module contains various parameters  of files handling and 
        glossary of some technical words.

.. warnings:: `_sensible` class is parser class. It parses of differents inputs 
                format recognized  by the software. Editing this class could affect 
                how others submodules could run. However, improving the code to become 
                less expensive is welcome.
                ...
        
        
Created on Sat Dec 12 16:21:10 2020

@author:KouaoLaurent alias @Daniel03

"""
import os, re ,warnings
import numpy as np 
from pycsamt.utils._csamtpylog import csamtpylog 
from pycsamt.utils.exceptions import pyCSAMTError_file_handling




class notion :
    """
    Singular class to explain CSAMT technical word in details. Also usefull for user to have info 
    about any scientific context is does Know. Just call the word directly in warnings to give an overview
    of why error occurs. It like a short documentation  using pyCSAMT software. 
    Used everywhere in the script . 
    
    """
    
    reference_frequency ='"Reference number " is The highest frequency with clean data.'
    
    TMA='Trimmed-Moving-Average filter : design to remove single-station offsets'\
        'while preserving broader scale changes.it is uded to estimate average apparent'\
        'resistivities at a single static-correction-reference frequency.'
        
    AVG='AVG is typical file from AMTAVG Zonge Engeneering software. It averages GDP CSAMT (CSAM) and Harmonic CSAMT (CSHA) raw data. Several files may be'\
        'created, including a log file (.LOG-file), listing file (.AL-file), average data files (.AD- and .AVG-file),'\
        ' plot file (.Z-file), and vector files (.Xnn-files).'
            
    FMLA= 'The FLMA filter estimates static-corrected apparent resistivities at a single reference frequency by' \
            'calculating a profile of average impedances along the length of the line. Use a fixed-length-moving-average'\
            'filter to estimate average apparent resistivities at a single static-correction-reference frequency.'
            
    Hanning_window = 'The Hanning window is a taper formed by using a weighted cosine.'\
        ' The Hanning was named for Julius von Hann, an Austrian meteorologist.'\
        ' It is also known as the Cosine Bell. Some authors prefer that it be called a'\
        ' Hann window, to help avoid confusion with the very similar Hamming window.'   
    
    weighted_Beta = 'The coefficient Betaj describe the way in which dipole responses adjacent to the sounding site x = Xk'\
                    'are weighted to emulate the shape of a Hanning window. For a given window width Wand center point Xko these weights'\
                    'can be adjusted so that the difference between the continuous Hanning window h( x - X k) and the synthesized version'\
                    'of it h(x. xd is minimized in a least-squares sense.'
    j= ''.join(['J-format is Magnetotelluric data file  written by A.G. Jones file format:',
                    ' The convention used is for RXY to represent the E-polarization (TE) mode,',
                    'and for RYX to represent the B-polarization mode. ', 'Consulting http://mtnet.info/docs/jformat.txt ',
                    'for more documentation. Some example file can be find at : http://mtnet.info/docs/jformat_example.txt '])
    
    occam_mesh_parameter_specs_info = ''.join(['Params specs is four lines for each layer.',
                                          'The four lines contain symbols which represent the value of four triangles in each mesh block.', 
                                          ' These four lines stand for the top, left, bottom, and right-hand triangles of each block in a layer.',
                                          'The character <?,Z,A|Y,0> indicates what resistivity value is to be assigned to the triangle.'
                                          ])                           
    occam_mesh_parameter_specs_meaning = {'ZZZZZZZZZZZZ': 'means that the triangle contains seawater', 
                                          '????????????': 'means that the triangle is a free parameter to be calculated',
                                          '000000000000': 'means that it contains air',
                                          'YYYAAAAAAAYYY' :'‘A’ through ‘Y’indicate fixed resistivity valuesas specified above.' }
    
    j_recordR ={'period': 'period in s, if negative then is frequency in Hz.',
                        'rho'   : 'apparent resistivity, if negative then rejected.',
                        'pha'   : 'phase in degrees',
                        'rhomax': 'rho + 1 standard error',
                        'rhomin': 'rho - 1 standard error',
                        'phamax': 'pha + 1 standard error',
                        'phamin': 'pha - 1 standard error',
                        'wrho'  : 'weight for rho, if <0 then rejected',
                        'wpha'  : 'weight for phase',
                        -999. : 'missing data marker'
            }
    j_recordZ ={ 'period': 'period in s, if negative then is frequency in Hz.',
                        'real'  : 'real part of the transfer function.',
                        'imag'  : 'imaginary part of the transfer function.',
                        'error' : 'standard error.',
                        'weight': 'weight (Note: some weights are in error!).',
                        -999. : 'missing data marker.',
                }
    
    utm_zone_dict_designator ={
                  'X': [72,84], 'W': [64,72], 'V':[56,64], 'U':[48,56],
                  'T': [40,48], 'S': [32,40], 'R':[24,32], 'Q':[16,24], 
                                      'P': [8,16],  'N':[0,8], 'M':[-8,0], 'L':[-16, 8], 
                                      'K':[-24,-16], 'J':[-32,-24], 'H':[-40,-32], 'G':[-48,-40], 
                                      'F': [-56,-48], 'E': [-64, -56], 'D': [-72,-64], 'C':[-80,-72],
                                      'Z':[-80,84]
                      }
        
class suit :
    """
    Singular class to easy manipulate word. used everywhere in the script to avoid redondancy. 
    
    """
    idle=[' ', 'nan', np.nan, '*','NaN','none', None, '**' ]
    longitude =['lon', 'Longitude', 'LONGITUDE', 'LON', 'LONG', 'longitude']
    latitude =['lat', 'Latitude', 'LATITUDE', 'LAT', 'Lat', 'latitude']
    easting =['e','east', 'EASTING', 'easting'] 
    northing =['n','north', 'NORTHING', 'northing'] 
    station =['dot', 'station', 'sta','st']
    elevation=['h', 'elev', 'elevation', 'ELEV', 'ELEVATION']
    topography=['topography']
    stn_separation=['station separation stn']
    azimuth=['azimuth']



class _sensitive:
    """
    .. note :: `sensitive class` . Please keep carefully the indices like 
        it's ranged. Better Ways to avoid redundancy in code during the program 
        design. Core of parser of each files except files from :ref:`module-geodrill`.
        Aims are:
        1. to check file . it is was the right file provided or not 
        2. to write file . to Zonge Avg_format or J-Format or EDI -format and else.
        3. to compute value . Indice are used for computation , set and get specific value.
    
    """
    _logger =csamtpylog.get_csamtpy_logger(__name__)
    
    _j=['>AZIMUTH', '>LATITUDE','>LONGITUDE','>ELEVATION',
       'RXX', 'RXY', 'RYX', 'RYY', 'RTE', 'RTM', 'RAV',
       'RDE', 'RZX', 'RZY', 'SXX', 'SXY', 'SYX', 'SYY', 'STE',
       'STM', 'SAV', 'SDE', 'SZX', 'SZY', 'ZXX', 'ZXY', 'ZYX', 
                   'ZYY', 'ZTE', 'ZTM', 'ZAV', 'ZDE', 'ZZX', 'ZZY', 'QXX',
                   'QXY', 'QYX', 'QYY', 'QTE', 'QTM', 'QAV', 'QDE', 'QZX',
                   'QZY', 'CXX', 'CXY', 'CYX', 'CYY', 'CTE', 'CTM', 'CAV', 
                   'CDE', 'CZX', 'CZY', 'TXX', 'TXY', 'TYX', 'TYY', 'TTE', 
                   'TTM', 'TAV', 'TDE', 'TZX', 'TZY' ,'ZXX', 'ZXY','ZYX']
    
    _avg=['skp','Station',' Freq','Comp',' Amps','Emag',
         'Ephz','Hmag','Hphz','Resistivity','Phase',
         '%Emag','sEphz','%Hmag','sHphz','%Rho','sPhz', 
                             'Tx.Amp','E.mag','E.phz','B.mag','B.phz','Z.mag',
                             'Z.phz','ARes.mag','SRes','E.wgt', 'B.wgt','E.%err',
                             'E.perr','B.%err','B.perr','Z.%err','Z.perr',
                             'ARes.%err']
    
    
    _edi =['>HEAD','>INFO', #Head=Infos-Freuency-Rhorot,Zrot and end blocks
        '>=DEFINEMEAS','>=MTSECT', '>FREQ ORDER=INC', #Definitions Measurments Blocks
        '>ZROT','>RHOROT','>!Comment','>FREQ','>HMEAS','>EMEAS',
        
              '>RHOXY','>PHSXY','>RHOXY.VAR','>PHSXY.VAR','>RHOXY.ERR','>PHSXY.ERR','>RHOXY.FIT','>PHSXY.FIT', #Apparents Resistivities  and Phase Blocks
              '>RHOYX','>PHSYX','>RHOYX.VAR','>PHSYX.VAR','>RHOYX.ERR','>PHSYX.ERR','>RHOYX.FIT','>PHSYX.FIT',
              '>RHOXX','>PHSXX','>RHOXX.VAR','>PHSXX.VAR','>RHOXX.ERR','>PHSXX.ERR','>RHOXX.FIT','>PHSXX.FIT',
              '>RHOYY','>PHSYY','>RHOYY.VAR','>PHSYY.VAR','>RHOYY.ERR','>PHSYY.ERR','>RHOYY.FIT','>PHSYY.FIT',
              '>FRHOXY','>FPHSXY','>FRHOXY.VAR','>FPHSXY.VAR','>FRHOXY.ERR','>FPHSXY.ERR','>FRHOXY.FIT','>FPHSXY.FIT', 
              '>FRHOXX','>FPHSXX','>FRHOXX.VAR','>FPHSXX.VAR','>FRHOXX.ERR','>FPHSXX.ERR','>FRHOXX.FIT','>FPHSXX.FIT',
                        '>TSERIESSECT', '>TSERIES', '>=SPECTRASECT', '>=EMAPSECT', '>=OTHERSECT', #Time series-Sepctra and EM/OTHERSECT
                              
                        '>ZXYR','>ZXYI','>ZXY.VAR', '>ZXYR.VAR','>ZXYI.VAR', '>ZXY.COV', #Impedance Data Blocks
                        '>ZYXR','>ZYXI','>ZYX.VAR', '>ZYXR.VAR','>ZYXI.VAR', '>ZYX.COV',
                        '>ZYYR','>ZYYI','>ZYY.VAR', '>ZYYR.VAR','>ZYYI.VAR', '>ZYY.COV',
                        '>ZXXR', '>ZXXI','>ZXXR.VAR','>ZXXI.VAR','>ZXX.VAR','>ZXX.COV',
                        '>FZXXR','>FZXXI','>FZXYR','>FZXYI', 

                                      '>RES1DXX', '>DEP1DXX', '>RES1DXY', '>DEP1DXY', #Continuous 1D inversion 
                                      '>RES1DYX', '>DEP1DYX', '>RES1DYY', '>DEP1DYY',
                                      '>FRES1DXX', '>FDEP1DXX', '>FRES1DXY', '>FDEP1DXY',
          
                        '>COH','>EPREDCOH','>HPREDCOH','>SIGAMP','>SIGNOISE', # Coherency and Signal Data Blocks
          
                        '>TIPMAG','>TIPPHS','TIPMAG.ERR', '>TIPMAG.FIT', '>TIPPHS.FIT', #Tipper Data blocks 
                        '>TXR.EXP', '>TXI.EXP', '>TXVAR.EXP','>TYR.EXP','>TYI.EXP','>TYVAR.EXP',
                        '>ZSTRIKE','>ZSKEW','>ZELLIP', '>TSTRIKE', '>TSKEW','>TELLIP', #Strike, Skew, and Ellipticity Data Blocks 
                        
                        '>FILWIDTH','>FILANGLE','>EQUIVLEN' , '>END'] #Spatial filter blocks
    
    _occam_startup =['Format', 'Description', 'Model File',
                    'Data File','Date/Time', 'Iterations to run','Target Misfit','Roughness Type','Diagonal Penalties',
                    'Stepsize Cut Count','Model Limits','Model Value Steps','Debug Level','Iteration',
                    'Lagrange Value','Roughness Value','Misfit Value','Misfit Reached','Param Count'] # end with file or startup 
    
    _occam_datafile =['FORMAT','TITLE','SITES','OFFSETS (M)','FREQUENCIES','DATA BLOCKS','DATUM','ERROR'] #data of simple file.or dat 
                     
    _occam_modelfile =['FORMAT', 'MODEL NAME','DESCRIPTION','MESH FILE','MESH TYPE','STATICS FILE','PREJUDICE FILE','BINDING OFFSET','NUM LAYERS'] # model of simple file 
                      
    _occam_meshfile =['?????????????','ZZZZZZZZZZZZ','YYYYYYYYYYYY','000000000000', 'AAAAAAAAAAAAA'] # end by mesh or simple file 
                            #Check the firstS layer (LIST NUMBER 0 135 52 0 0 2) SPECIFIED the catacteristik of mesh
                            #only difference between the startup is at the end , the startup file have the same number so 
                            # the iter is not the same case , number change .
                            #response file , check if the range or the lenght == 7. 
    _occam_logfile =['STARTING R.M.S', ' ** ITERATION','TOFMU: MISFIT',
                    'AND IS','USING','STEPSIZE IS','ROUGHNESS IS']          #or end with Logfile.
    
    _stn=['""dot','""e','""h', '""n','""sta','""len','""east',
         '""north','""lat','""long','""elev','""azim','""stn']
    
    
    @classmethod
    def which_file(cls, filename=None, deep= True): 
        """
        Which file is class method . List of files are the typical files able
        to read by pyCSAMT softwares.
        Sensitive class method. 
        
        Parameters
        ----------
            **filename** :str 
                            corresponding file to read , pathLike 
            **deep** : bool , 
                    control reading : False for just control the extension file ,
                    not opening file . True  control in deeper file and
                     find which file were inputted.
                
        Returns  
        ---------
            str 
               FileType could be [`avg` | `j` | `edi` | `resp` | `mesh` | `occamdat` |
                    `stn` | `model` | `iter` | `logfile` | `startup`]
        
       
        List of files read by pyCSAMT :
        
        ==============  =======================================================
        CodeFile                        Description
        ==============  =======================================================
        *avg*           Zonge Engineering file Plainty file of ASTATIC file. 
        *j*             A.G .Jonhson J=Format file. 
        *edi*           SEG (Society of Exploration Geophysics) Electrical
                        Data Interchange file (SEG-EDI) 
        *stn*           Zonge Engineering station file 
        *occamdat*      deGroot-Hedlin, C., and S. Constable, Occam file. 
        *mesh*          Constable, S. C., R. L. Parker, and C. G. Constable 
                        mesh file 
        *model*         Occam Model file  
        *startup*       Occam startup file  
        *iter*          Occam iteration file get after Inversion  
        *resp*          Occam response file , Get after inversion 
        *logfile*       Occam Logfile, Get after inverson (Inversion file )
        ==============  ======================================================= 
        
       
        :Example:
       
            >>> files = ['K1_exp.bln','LCS01.avg' ,'LCS01_2_to_1.avg', 'K1.stn',
            ...            'csi000.dat','csa250.edi','LogFile.logfile',
            ...                 'Occam2DMesh','Occam2DModel', 'OccamDataFile.dat',
            ...            'S00_ss.edi', 'Startup','RESP13.resp', 
            ...                 'ITER02.iter']
            >>>  for ii in files : 
            >>>      path =  os.path.join(os.environ["pyCSAMT"], 
            ...                                  'csamtpy','data', ii)
            ...       try : 
            ...         print(_sensitive.which_file(path,deep=True))
            ...       except :pass 
        """
        
        _code = {'edi':cls._edi , 'j': cls._j, 'startup':cls._occam_startup, 
                 'occamdat':cls._occam_datafile,'model': cls._occam_modelfile,
                 'mesh': cls._occam_meshfile, 'iter': cls._occam_startup, 
                 'avg': cls._avg,  'stn': cls._stn,'logfile':cls._occam_logfile,
                 'resp': ['none','none'],
                  }
        
        
        _filename = filename
        
        if _filename is None : pyCSAMTError_file_handling('Error file: NoneType can not be computed. Check your right path.') 
        if _filename is not None : 
            if os.path.basename(_filename).split('.')[-1].lower() in list(_code.keys()): #code file recognized
                _flag = os.path.basename(_filename).split('.')[-1].lower()
                if deep ==False :return _flag
            else : _flag = 0 #file not recognized 

        # Open the file now 
        with open (_filename , 'r', encoding ='utf8') as _f : 
                _data_lines =_f.readlines()
                
        if _flag !=0 :  
                                   # file extension has been recognized on dict code keys. 
            cls._logger.info('Reading %s file' % os.path.basename(_filename))
            
            _value_code = _code[_flag]
            
            if _flag =='edi':                   #avoid expensive function then controle 2thing before continue 
                if _value_code [0] not in  _data_lines[0] :
                        warnings.warn('File probably is incorrect. Please consult EDI_Main file components : {0}'.\
                                      format(''.join(cls._edi)))
                        raise pyCSAMTError_file_handling("Wrong SEG-EDI file. It'seems -File provided doesnt include '>HEAD. Please provide the right EDI-FORMAT")
                elif _value_code[-1] not in _data_lines[-1] :
                    warnings.warn('File probably incorrect. Please consult EDI_Main file composents : {0}'.\
                                  format(''.join(cls._edi)))
                    raise pyCSAMTError_file_handling("Wrong SEG-EDI file. It'seems -File provided doesnt include '>END. Please provide the right EDI-FORMAT")
            
            elif _flag =='resp' :               # control quick response file
                                                #check if the len of resp= 7 
                if len( _data_lines[0].strip().split()) !=7 : 
                    raise pyCSAMTError_file_handling("Wrong OCCAM-RESPONSE file. Please provide the right response file.")
                else : return _flag 
                
            elif _flag =='avg': 
                _reason = cls.validate_avg(avg_data_lines=_data_lines)[0]
                if _reason not in ['yesAST', 'yes'] : 
                    warnings('File <{0}> does not match Zonge AVGfile. Get info to {1}'.format(os.path.basename(_filename), notion.AVG))
                    raise pyCSAMTError_file_handling("Wrong <{0}>Zonge AVG-file. Please provide the right file.".format(os.path.basename(_filename)))
                return _flag
            
            elif _flag =='mesh': 
               try :[float(ss) for ss in _data_lines[1].strip().split()] #check the first occam layer 
               except :
                   warnings.warn('Wrong OCCAM-MESH file.Get more info on that site :https://marineemlab.ucsd.edu/Projects/Occam/index.html')
                   raise pyCSAMTError_file_handling("Wrong OCCAM-MESH <%s>file .We don't find the characteristic of thefirst layer."\
                                                    "Please provide the right OCCAM meshfile." % os.path.basename(_filename)) 
               for _meshcodec in _value_code: 
                   if _meshcodec in _data_lines[-1]: return _flag 
                   else :
                       warnings.warn('Trouble occurs while reading the "OCCAM-params specs.More details in :{0}.\n {1}'.
                                     format(notion.occam_mesh_parameter_specs_info, notion.occam_mesh_parameter_specs_meaning))
                       raise pyCSAMTError_file_handling('We do not find a "Parameter specs" in your <{0}>'.format(os.path.basename(_filename)))
                   
                   
            #once both thing has already been controled . then continue . avoid avg because already has deep controlled 
            # with special static method then pass 
            _ccounter =0 # threshold control is match to 7 except stn.
            for _codec in _value_code : # loop the dict value of code  #loop now all the
                if _flag not in ['avg', 'resp','mesh']: 
                    for num , _data_item in enumerate(_data_lines):
                        if _codec in _data_item :
                            _ccounter += 1
                    if _flag =='stn': 
                        if _ccounter >= 3 : return _flag 
                    elif _flag =='j':
                        if _ccounter >=5 : return _flag
                    if _ccounter>=7 :return _flag 
            if _ccounter <7 :raise pyCSAMTError_file_handling('ERROR occurs while reading file !<{0}>. '\
                                                              'File doesnt not match the typical <{1}> Format.'.format(os.path.basename(_filename), _flag))  
        
        elif _flag == 0 : #file not recognized in _codekeys , 
                             #then try to read and to find which file mastch the best.
            _ccounter=0
            for _keycode, _value_code in _code.items(): 
                for num, valueitem in enumerate(_data_lines) :
                    for codec  in _value_code : #read value in code 
                        if codec.lower() in valueitem.lower(): 
                            _ccounter +=1
                    if _keycode =='j' :
                        if _ccounter >=4 :return _keycode
                    if _ccounter >= 7 :
                        if _keycode =='startup':
                            if all([float(val) for val in _data_lines[len(_value_code)+1].strip().split()]) == False :
                                return 'iter'
                            else :return _keycode
                        return _keycode
                    
            #Now we assune that the file must be the response file then read according response control 
            try :
                resp0 =[float(ss) for ss in _data_lines [0]]
                resp_end =[float(ss) for ss in _data_lines[-1]]
                resp_mid = [float(ss) for ss in _data_lines [int(len(_data_lines)/2)]] # get the middele of file and try float converter 
                assert (len(resp0) == len(resp_end)) and (len(resp_end)==len(resp_mid)),\
                    pyCSAMTError_file_handling('We presume that  RESP <{0}> file doesnt not match a OCCAM-RESPONSE file.'.format(os.path.basename(_filename)))
            except :
                
                message =''.join('Error reading File. pyCSAMT can read only <{0}> Format file.', 
                             'Format file. Program can not find the type of that file : <{1}>.', 
                             'You may use script <rewriteXXX> of  pyCSAMT to try to rewrite file', 
                             'either to SEG-EDI |j|STN|AVG| or OCCAM{DAT, RESP,MODEL,ITER} files.')
                cls._logger.error(message.format('|'.join(list(_code.keys())),os.path.basename(_filename) ))
                warnings.warn(message.format('|'.join(list(_code.keys())),os.path.basename(_filename) ))

                return 
            
            return 'resp'


                 
    @staticmethod 
    def validate_avg(avg_data_lines):
        """
        Core function  to validate avg file .
        
        :param avg_data_lines: list of avgfile
        :type avg_data_lines: list 
        
        :returns: 'yesAST' or 'yes' where  'yesAST' is  Astatic file 
                    and  'yes'  is plainty avg file (the main file)  
        :rtype: str 
        
        :returns: item, spliting the headAvg components strutured by file.
        :rtype: list 
        
        :Example:
            
            >>> from csamtpy.etc.infos.Infos inmport _sensistive as SB
            >>> path =  os.path.join(os.environ["pyCSAMT"], 
            ...                                  'csamtpy','data', LCS01_2_to_1.avg)
            ... with open (path, 'r', encoding ='utf8') as f : 
            ...   datalines = f.readlines()
            ... ss =SB._sensitive.validate_avg(avg_data_lines=datalines)
        """
        
        num=[0]
        for jj, code in enumerate(['ASTATIC','$Stn','$Survey', '$Tx','$Rx',
                                   'ARes.mag','SRes','Z.phz','Tx.Amp','ARes.%err']):
            for items in avg_data_lines :
                if items.find(code)> 0:num.append(1) 
                if jj >=5 and(sum(num) >=3):
                    if items.find(code)> 0: item =items.strip().split(',')
                    return ('yesAST', item)
            if jj >=5 and (sum(num)<3):break

        for jj, code in enumerate(['AMTAVG','ASPACE' ,'XMTR' ,'skp' ,'\\-',
                                    'Resistivity','Phase','Comp',' Amps','%Rho']):
            for items in avg_data_lines :
                if items.find(code)> 0:num.append(1) 
                if jj >=4 and(sum(num) >=3):
                    if items.find(code)> 0:
                        item =items.strip().split()
                        return ('yes', item) 
                    
        if sum(num) ==0 : raise pyCSAMTError_file_handling(
                'Data format provided are not typically issued from Zonge AVG format .')
           
 
    
    


# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  

    .Module written,  inspired on  the J.R Peacob model (MTpy)
    .looging file was tst only on .yaml file.  
    . see load_configure_set_logfile (only for .yaml file )
    
Created on Sun Sep 27 19:29:59 2020

@author: KouaoLaurent @Daniel03 

A more Pythonic way of logging:
Define a class MtPyLog to wrap the python logging module;
Use a (optional) configuration file (yaml, ini, json) to configure the logging,
It will return a logger object with the user-provided config setting.
see also: http://www.cdotson.com/2015/11/python-logging-best-practices/


"""
import os 
import yaml
import logging 
import logging.config
import inspect 


class csamtpylog:
    """
    Field to logs csamtpy module Files  in order to tracks all 
    exceptions
    
    """
    
    @staticmethod
    def load_configure (path2configure =None, OwnloggerBaseConf=False) :
        """
        configure/setup the logging according to the input configfile

        :param configfile: .yml, .ini, .conf, .json, .yaml.
        Its default is the logging.yml located in the same dir as this module.
        It can be modified to use env variables to search for a log config file.
        """
        
        configfile=path2configure
        
        if configfile is None or configfile == "":
            if OwnloggerBaseConf ==False :
                logging.basicConfig()
            else :
                csamtpylog().set_logger_output()
            
        elif configfile.endswith(".yaml") or configfile.endswith(".yml") :
            this_module_path=os.path.abspath(__file__)
            # print(this_module_path)
            
            logging.info ("this module is : %s", this_module_path)
            yaml_path=os.path.join(os.path.dirname(this_module_path),
                                   configfile)
            
            logging.info('Effective yaml configuration file %s', yaml_path)

            if os.path.exists(yaml_path) :
                with open (yaml_path,"rt") as f :
                    config=yaml.safe_load(f.read())
                logging.config.dictConfig(config)
            else :
                logging.exception(
                    "the config yaml file %s does not exist?", yaml_path)
                
        elif configfile.endswith(".conf") or configfile.endswith(".ini") :
            logging.config.fileConfig(configfile,
                                     disable_existing_loggers=False)
            
        elif configfile.endswith(".json") :
            pass 
        else :
            logging.exception("logging configuration file %s is not supported" %
                configfile)
            
    
    @staticmethod        
    def get_csamtpy_logger(loggername=''):
        """
        create a named logger (try different)
        :param loggername: the name (key) of the logger object in this Python interpreter.
        :return:
        """
        logger =logging.getLogger(loggername)
        csamtpylog.load_configure() #set configuration 

        return logger
    
    
    @staticmethod
    def set_logger_output(logfilename="myapplog.log", 
                      date_format='%m %d %Y %I:%M%S %p',
                      filemode="w",format_="%(asctime)s - %(name)s - %(levelname)s - %(message)s",#'%s(asctime)s - %(message)s',
                       level=logging.DEBUG ):
        """
        
        Parameters
        ----------
        * logfilename : str, optional
                        logfilename output. The default is "LogFileTest.log".
        * date_format :str, 
                        format of date . consult module time. The default is '%m %d %Y %I:%M%S %p'.
        * filemode : str, 
                        mode to write the out[ut file. Must be ["a","w", "wt',"wt"].
                                                    The default is "a".
        * format_ : str,
                        format of date and time. Must consult the module of datetime.
                        The default is '%s(asctime)s %(message)s'.
        * level : object, optional
                    logging object.  Must be C.E.W.I.D :
                        >>> logging.CRITICAL
                        >>> Logging.ERROR
                        >>> logging.WARNING
                        >>> logging.INFO
                        >>> logging.DEBUG
                    The default is logging.DEBUG.

        Returns : NoneType object
        -------
        Procedure staticFunctions

        """
        # #create logger 
        logger=csamtpylog.get_csamtpy_logger(__name__) # logger=logging.getLogger(__name__)
        logger.setLevel(level)
        
        # #create console handler  and set the level to DEBUG 
        ch=logging.StreamHandler()
        ch.setLevel(level) # level=logging.DEBUG
        
        #create formatter :
        formatter =logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        
        #add Formatter to ch 
        ch.setFormatter(formatter)
        
        #add ch to logger 
        logger.addHandler(ch)
        
        
        # logging.basicConfig(filename="test.log", level=logging.DEBUG)
 
        logging.basicConfig(filename=logfilename, format=format_,
                            datefmt=date_format,
                            filemode=filemode, level=level)
        logging.debug("debug message")
        logging.info("info message ")
        logging.warning("warn message")
        logging.error("error message")
        logging.critical("critical message")
        
    @staticmethod
    def load_configure_set_logfile (path2configfile=None): # loggername =None, setLevel=Name
        """
        configure/setup the logging according to the input configure .yaml file.

        :param configfile: .yml, or add ownUsersConfigYamlfile (*.yml) 
        Its default is the logging.yml located in logfiles folder 
        It can be modified to use env variables to search for a log config file.
        
        """
        
        ownUserLogger="main_logging_configfile.yml"
        if path2configfile is None :
            env_var=os.environ['pyCSAMT']
            path2configfile =os.path.join( env_var, 'csamtpy','geoDrill','_logfiles',
                ownUserLogger)
            

        elif path2configfile is not None :
            if os.path.isdir(os.path.dirname(path2configfile)):
                if path2configfile.endswith('.yml') or path2configfile.endswith('.yaml'):
                    
                    logging.info('Effective yaml configuration file :%s', path2configfile)
                else :
                    logging.exception('File provided {%s}, is not a .yaml config file !'%os.path.basename(path2configfile))
            else :
                
                logging.exception ('Wrong path to .yaml config file.')
        
        yaml_path=path2configfile
        os.chdir(os.path.dirname(yaml_path))
        if os.path.exists(yaml_path) :
            with open (yaml_path,"rt") as f :
                config=yaml.safe_load(f.read())
            logging.config.dictConfig(config)
        else :
            logging.exception(
                "the config yaml file %s does not exist?", yaml_path) 
                
                    

def test_yaml_configfile(yamlfile="logging.yml"):
    
    this_func_name = inspect.getframeinfo(inspect.currentframe())[2]
    # print("this_func_name:\n",this_func_name)
    # print(("starting", this_func_name))
    
    UsersOwnConfigFile = yamlfile
    csamtpylog.load_configure(UsersOwnConfigFile)
    # print(myobj)
    logger = csamtpylog.get_csamtpy_logger(__name__)
    
    print((logger, id(logger), logger.name, logger.level, logger.handlers))
    
    # 4 use the named-logger
    logger.debug(this_func_name + ' __debug message')
    logger.info(this_func_name + ' __info message')
    logger.warn(this_func_name + ' __warn message')
    logger.error(this_func_name + ' __error message')
    logger.critical(this_func_name + ' __critical message')

    print(("End of: ", this_func_name))
    
def test_config_configfile(configfile="logging.conf"): 
    
    this_func_name=inspect.getframeinfo(inspect.currentframe())[1]
    
    print(('starting', this_func_name))
    UsersOwnConfigFile =configfile
    
    csamtpylog.load_configure(UsersOwnConfigFile)
    
    logger=csamtpylog.get_csamtpy_logger(__name__)
    
    print((logger, id(logger), logger.name, logger.level, logger.handlers))
    
    logger.debug(this_func_name + ' __debug message')
    logger.info(this_func_name + ' __info message')
    logger.warn(this_func_name + ' __warn message')
    logger.error(this_func_name + ' __error message')
    logger.critical(this_func_name + ' __critical message')
    
    print(("End of :", this_func_name))
    
    
def test_none_configfile():
    
    this_func_name =inspect.getframeinfo(inspect.currentframe())[2]
    
    print(('Starting of :', this_func_name))
    
    UsersOwnConfigFile =''
    
    csamtpylog.load_configure(UsersOwnConfigFile)
    
    logger=csamtpylog.get_csamtpy_logger(__name__)
    
    print((logger, id(logger), logger.name, logger.level, logger.handlers))
    
    logger.debug(this_func_name + ' __debug message')
    logger.info(this_func_name + ' __info message')
    logger.warning(this_func_name + ' __warn message')
    logger.error(this_func_name + ' __error message')
    logger.critical(this_func_name + ' __critical message')
    
    print(("End of :", this_func_name))
    
    
        
def test_jsaon_configfile():
    pass 



if __name__=="__main__":

    csamtpylog.load_configure_set_logfile()
    logger=csamtpylog.get_csamtpy_logger(__name__)
    # this_func_name=inspect.getframeinfo(inspect.currentframe())[1]
    # test_yaml_configfile(ownUserLogger)
    logger.debug(' __debug message')
    logger.info(' __info message')
    logger.warning(' __warn message')
    logger.error(' __error message')
    logger.critical(' __critical message')
    
   
    
    
    
    
    
    
    
    
            
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        



# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================

.. _module-ZCalculator::`pycsamt.ff.processing.zcalculator`
    :synopsis:  helper functions special calculator 

Created on Thu Dec  3 16:44:29 2020

@author: @Daniel03
"""
import os,warnings 
import numpy as np 
from pycsamt.etc import infos
from pycsamt.utils import exceptions as CSex
from pycsamt.utils import func_utils as func
from pycsamt.utils.decorator import deprecated, redirect_cls_or_func
from pycsamt.utils._csamtpylog import csamtpylog 

_logger =csamtpylog.get_csamtpy_logger(__name__)
try:
    import scipy

    scipy_version = [int(ss) for ss in scipy.__version__.split('.')]
    if scipy_version[0] == 0:
        if scipy_version[1] < 14:
            warnings.warn('Note: need scipy version 0.14.0 or higher or interpolation '
                          'might not work.', ImportWarning)
            _logger.warning('Note: need scipy version 0.14.0 or higher or interpolation '
                            'might not work.')
    import scipy.interpolate as spi

    interp_import = True

except ImportError: 
    warnings.warn('Could not find scipy.interpolate, cannot use method interpolate'
                  'check installation you can get scipy from scipy.org.')
    _logger.warning('Could not find scipy.interpolate, cannot use method interpolate'
                    'check installation you can get scipy from scipy.org.')
    interp_import = False

############################# END MODULE IMPORT ############################################

mu0= 4* np.pi * 1e-7 
phase_referencing = 2 * np.pi 

def mag_avg (mag_array , A_spacing, Txcur): 
    """
    RAW, E-, or H-field magnitude values)for each frequency
        units : mV/Km*
        
    :param mag_array:  magnnitude value for each data block
    :type mag_array: np.array (ndarray, 1),

    :param a_spacing: dipole length 
    :type a_spacing: float 
    
    :param txcur:  curv coil, transmitter length 
    :type txcurv: float

    :returns: averaged data of magnitude data Block
    :rtype mag_avg: np.ndarray 
    
    """

    
    if type(mag_array) is list : 
        mag_avg =np.array(mag_array) 
    try : 
        mag_avg =np.array([float(ii) for ii in mag_array])
    except : 
        raise CSex.pyCSAMTError_inputarguments("Input first arguments is wrong. must"\
                                               "be a list of ndarray of numbers.")
        
    mag_avg =mag_array.mean()
    En =mag_avg /(A_spacing *  Txcur)
    
    return mag_avg, En 


def phz_avg (phz_array, to_degree): 
    """
    E-, H-field, or Impedance Phase values unit in  mrad
    
    :param phz_array: array of phase values in mrad 
    :type phz_array: array_like 
    
    :param to_degree: ascertain convertion to degree 
    :type to_degree: bool 

    """
    if type (phz_array) is list : 
        phz_array =np.array(phz_array)
        
    try : 
        phz_array =[float(kk) for kk in phz_array ]
    except : 
        raise CSex.pyCSAMTError_inputarguments("input argument must be a number in array or list.")
    
    if to_degree : return (phz_array.mean()/1e3) *180/np.pi, (phz_array/1000) *180/np.pi
    
    return phz_array.mean()
    
def param_rho (rho_array):
    """
    Parameter Average RHO (RHOp), RHO are from each data block, unit ( ohm.m)
    
    :param rho_array: array of resistivity values 
    :type rho_array: array_like 
    """
    if type(rho_array) is list : rho_array=np.array(rho_array)
    try : 
        phz_array =np.array([float(kk) for kk in rho_array ])
    except : 
        raise CSex.pyCSAMTError_inputarguments('Error inputs arguments . must be list or'\
                                               'ndarray of numbers.')
    return  phz_array.mean()

def param_phz (pphz_array, to_degree =False): 
    """
    PHZn are from each data block ,  units in  mrad 
       
    :param pphz_array: average parameters phase for data blocs.
    :type pphz_array: array_like
    
    :param to_degree: ascertain convertion to degree 
    :type to_degree: bool 
    """
    if type(pphz_array) is list : pphz_array=np.array(pphz_array)
    try : 
        pphz_array =np.array([float(kk) for kk in pphz_array ])
    except : 
        raise CSex.pyCSAMTError_inputarguments('Error inputs arguments. must be list or'\
                                               'ndarray of numbers.')
    if to_degree : return (pphz_array.mean()/1000) *180/np.pi, (pphz_array/1000) *180/np.pi
    
    return  pphz_array.mean()

def comp_rho (mag_E_field, mag_H_field, freq_array, A_spacing, Txcurr ):
    """
    Function to compute component average 
    unit in in ohm.m 

    :param mag_E_field: magnitude of E-filed , averaged 
    :type mag_E_field: np.ndarray(ndarray,1)
  
    :param mag_H_field: magnitude of H-Field ,  averaged 
    :type mag_H_field: np.ndarray(ndarray,1)
    
    :param freq_array:  frequency of station field
    :type freq_array: np.ndarray(ndarray,1)
    
    :param  A_spacing: step_between station
    :type A_spacing: np.float
    
    :param Txcurr:  distance of coil in meter 
    :type Txcurr: np.float,

    :returns: comp_rho , component averaged rho.
    :rtype: np.ndarray 
        
    """
    args =[mag_E_field, mag_H_field, freq_array, A_spacing, Txcurr]
    
    for a_args in args : 
        if type (a_args) is not np.ndarray :
            try :
                a_args =np.array([ float(ss) for ss in a_args ])
            except : 
                raise CSex.pyCSAMTError_inputarguments("Elemts composed of each array must be a number.")
        
    # if comp_rho.__code__.co_argcount ==5 : 
    #     if mag_E_field.ndim !=1 or  mag_H_field.ndim !=1 or freq_array.ndim !=1 : 
    #         raise CSex.pyCSAMTError_inputarguments ("dimension of Input arguments must me equal to 1.")
    
    Emag =mag_E_field /(A_spacing * Txcurr)
    Hmag =mag_H_field /Txcurr 
    comp_rho =(Emag/Hmag)**2 * (1/5*freq_array)*np.power(10,6)
    
    return comp_rho

def comp_phz (comphz_array, units ='deg'):
    
    """
    PHZc are from each data block, units in rad 

    :param comphz_array: average parameters phase for data blocs.
    :type comphz_array: float 
    
    :returns: component phase averaged.
    :rtype: component phase averaged.
    
    :Example:
        
        >>> path =  os.path.join(os.environ["pyCSAMT"], 
        ...              'csamtpy','data', 'K1.AVG')
        >>> from pycsamt.core import avg  
        >>> phs_obj =avg.Phase(path)
        >>> phs_obj.loc['S00']
        >>> value, ss = comp_phz(comphz_array=phs_obj.loc[
        ...            'S00'], to_degree=True)
        ... print(value)
    """
    
    if type(comphz_array) is list : comphz_array=np.array(comphz_array)
    
    if comphz_array.dtype not in ["float", "int"] : 
        raise CSex.pyCSAMTError_inputarguments("Type provided is wrong ! number must be float")  

    try : 
        comphz_array=np.array([float(kk) for kk in comphz_array ])
    except : 
        raise CSex.pyCSAMTError_inputarguments('Error inputs arguments. must be list or'\
                                               'ndarray of numbers.')
    
            
    if units =='deg' : return comphz_array.mean() *180/np.pi, comphz_array *180/np.pi
    
    return  comphz_array.mean()


def compute_components_Z_Phz(magn_E_field , magn_H_field, phz_E_field,
               phz_H_field, freq_value, **kwargs):
    """
    Function to compute all components  derived from Impedance Z. 
    user can  enter specifik units in kwargs arguments . program will compute and converts value 
    automatically.
    
    Parameters
    ----------
        * magn_E_field : np.ndarray 
            E_.field magnitude (ndarray,1) in  microV/KM*A
            
        * magn_H_field : np.ndarray 
             H_.field magnitude (ndarray,1)in  mGammas/A or picoTesla/A
             
        * phz_E_field : np.ndarray 
            E_field phase (ndarray, 1) in  mrad 
            
        * phz_H_field : np.ndarray 
             H_field phase (ndarray,1) in  mrad.
             
        * freq_value : np.ndarray 
             Frequency at which data was measured(ndarray,1)in  Hz
             
        * kwargs : str 
             units conversion.
        
    Raises
    ------
        CSex.pyCSAMTError_z(), 
            Exceptions if units entered by the user doesnt match or are messy.

    Returns
    -------
        rho: ndarray 
            Cagnard resistivity calculation. ohm.m 
        phz: ndarray 
             Impedance phase value.
        Zij: ndarray  
             Impedance Tensor value.
        Zreal: ndarray 
            Value of Real part of impedance Tensor.
        Zimag: float
             Value of Imaginary part of impedance Tensor.
        Zreal_imag: ndarray , complex 
            Complex value of impedance Tensor.
         
    :Example: 
        
        >>> from pycsamt.core import avg 
        >>> path =  os.path.join(os.environ["pyCSAMT"], 
        ...              data', 'avg', 'K1.AVG')
        >>> emag_ob = avg.Emag(path)
        >>> hmag_obj = avg.Hmag(path)
        >>> ephz_obj = avg.Ephz(path)
        >>> hphz_obj = avg.Hphz(path)
        >>> freq_obj =avg.Frequency(path)
        >>> station_name ='S00'
        >>> rho, phz, Z, real, imag, comp =compute_components_Z_Phz( 
        ...    magn_E_field=emag_ob.loc[station_name], 
        ...                            magn_H_field =hmag_obj.loc[station_name], 
        ...                            phz_E_field =ephz_obj.loc[station_name], 
        ...                            phz_H_field=hphz_obj.loc[station_name], 
        ...                            freq_value=freq_obj.loc[station_name])
        ... rho
    """
    
    units_E_field =kwargs.pop('unit_E_field', 'microV/km*A')
    units_H_field =kwargs.pop('unit_E_field', 'mGamma/A')
    unit_phz=kwargs.pop('unit_phase', 'mrad')
    unit_freq=kwargs.pop('unit', 'Hz')
    
    #units converter
    units ={units_E_field: {'microv/km*a':1/np.power(10,6),
                            'nv/ma':1/np.power(10,6),
                           'mmv/km*a':1/np.power(10,3),
                           'v/km*a':np.power(10,0)
                   },               # SI : --> E-Field  V/KM*A
            units_H_field : {'mgamma/a':1/np.power(10,6),
                             'pt/a':1/np.power(10,6),
                            'picotesla/a':1/np.power(10,6),
                            'gamma/km*a':np.power(10,3)
                   }, #SI : --> KGamma/A or microTesla/A 
            unit_phz:{'rad':np.power(10,0), 
                      'mrad': 1/np.power(10,3),
                      'deg': np.pi/180
                  },        # S.I : angle  on radians 
            unit_freq:{"hz":np.power(10,0), 
                         'rad/s': 0.1592,
                         'mrad/s':0.1592/np.power(10,3)} # on Hz 
            }
    #check units 
    for keys ,values in units.items():
        if keys.lower() == units_E_field.lower():
            if keys.lower() in values.keys():
                magn_E_field = magn_E_field * values[keys.lower()]
            else : 
                raise CSex.pyCSAMTError_Emag("Input units is not valid. try : {0}"\
                                                     " ".format(list(values.keys())))
        if keys.lower() ==units_H_field.lower() : 
            if keys.lower() in values.keys():
                magn_H_field= magn_H_field* values[keys.lower()]
            else : 
                raise CSex.pyCSAMTError_Emag("Input units is not valid. try : {0}"\
                                                 .format(list(values.keys()))) 
                    
        if keys.lower() ==unit_phz.lower(): 
            if keys.lower() in values.keys():
                
                phz_E_field = phz_E_field* values[keys.lower()]
                phz_H_field = phz_H_field  * values[keys.lower()]
            else : 
                raise CSex.pyCSAMTError_Emag("Input units is not valid. try : {0}"\
                                                 .format(list(values.keys()))) 
        if keys.lower() ==unit_freq.lower(): 
            if keys.lower() in values.keys():
                freq_value= freq_value * values[keys.lower()]
            else : 
                raise CSex.pyCSAMTError_Emag("Input units is not valid. try : {0}"\
                                                 .format(list(values.keys()))) 
        
    #--> component real part and imag 
    Zij =magn_E_field /magn_H_field
    phz=phz_E_field - phz_H_field
    Zreal, Zimag  =Zij * np.cos(phz), Zij * np.sin(phz)
    # Zreal_imag= np.complex(real=Zreal, imag=Zimag)
    Zreal_imag =np.complex(real=Zreal , imag=Zimag) 
    rho =(1/(5*freq_value))* Zij**2
    
    return rho , phz,  Zij, Zreal, Zimag, Zreal_imag


def z_error2r_phi_error(z_real, z_imag, error):
    """
    Error estimation from rectangular to polar coordinates.
    By standard error propagation, relative error in resistivity is 
    2*relative error in z amplitude. 
    Uncertainty in phase (in degrees) is computed by defining a circle around 
    the z vector in the complex plane. The uncertainty is the absolute angle
    between the vector to (x,y) and the vector between the origin and the
    tangent to the circle.
 
    :param z_real: real component of z (real number or array)
    :type z_real: float 
    
    :param z_imag:  imaginary component of z (real number or array)
    :type z_imag: complex
    
    :param error: absolute error in z (real number or array)
    :type error: float
    
    :returns: containers of relative error in resistivity, absolute error in phase
    :rtupe: tuple
    
    """
        
    z_amp = np.abs(z_real + 1j*z_imag)

    z_rel_err = error/z_amp
    
    res_rel_err = 2.*z_rel_err
    
    #if the relative error of the amplitude is >=100% that means that the relative 
    #error of the resistivity is 200% - that is then equivalent to an uncertainty 
    #in the phase angle of 90 degrees:
    if np.iterable(z_real):
        phi_err = np.degrees(np.arctan(z_rel_err))   
        phi_err[res_rel_err > 1.] = 90.
        
    else:
        if res_rel_err > 1.:
            phi_err = 90
        else:
            phi_err = np.degrees(np.arctan(z_rel_err))    
    
    
    return res_rel_err, phi_err

def rhophi2z ( phase, freq , resistivity=None,  z_array= None):
    """
    Function to compute z , real part and imag part . 

    :param phase: phase angles array in radians
    :type phase: ndarray 
     
    :param freq: frequencies array in Hz
    :type freq: array_like
    
    :param resistivity:  rho array in ohm.m
    :type resistivity: array_like
    
    :param z_array: impedance z array in V/m 
    :type z_array: array_like 

    :returns: z_abs , absolute value of zz
    :rtype: float 
    
    :returns: z_real, real part of complex number
    :rtype: float 
    
    :returns: z_imag, imaginary part of zz
    :rtype: complex 
    
    :returns: ndarray 
    :rtype: zz, array of z_abs, z_imag, z_real 
    """
    if (type(phase) is not np.ndarray) or (type(freq) is not np.ndarray) :
        phase, freq = np.array(phase), np.array(freq)
    
    if phase.dtype not in ['float', 'int'] or freq.dtype not in ['float', 'int']:
        try : 
            phase =np.array([float(ss) for ss in phase ])
            freq =np.array([float(f) for f in freq ])
        except : ValueError('must be flost number of integer . not str !')
    
    try :    
        if resistivity is not None : resistivity =np.array([float(res) for res in resistivity])
        if z_array is not None : z_array =np.array([float(z) for z in z_array])
    except : ValueError('Arguments number must be float or int')
    
    if freq.size != phase.size and (phase.size != resistivity.size or phase.size != z_array.size):
        raise 'Arrays nust get the same size.'
    
    
    if resistivity is not None : z_abs= np.sqrt(0.2 * freq * resistivity)
    if resistivity is  None and z_array is not None : z_abs = z_array 
    
    
    z_real, z_imag  = z_abs * np.cos (phase), z_abs * np.sin(phase)
    # zz =np.complex(z_real , z_imag)
    zz=z_real + z_imag *1j
    
    return z_abs, z_real, z_imag ,zz


def get_reffreq_index(freq_array, reffreq_value): 
    """ 
    Get the index of reference index. From this index ,All array will filter data at this reffreq
    value . 
    :param freq_array: array of frequency values
    :type freq_array: array_like 
    
    :param reffreq_value:  value of frequency at clean data 
    :type reffreq_value: float, int 
    """
    freq_array=np.array([float(ss) for ss in freq_array])
    if float(reffreq_value) not in freq_array : 
        raise CSex.pyCSAMTError_frequency('Reference frequency must be a value of frequency array.')

    for ii, freq in enumerate(freq_array): 
        if freq == reffreq_value:
            index_rf = ii
            break
    return index_rf

def interpolate_sets (array_to, fill_value =None , array_size=None ):
    """
    Function to interpolate data contain of multiple nan values. 
    """
    if array_size is None : array_size == array_to.size 
    
    xx_array_to = np.arange(array_size)
    
    ff = spi.interp1d(x=xx_array_to, y=array_to, fill_value=fill_value)
    # new_array = ff(xx_array_to)
    return ff(xx_array_to)

def  _interpolate_array_fromreffreq(stationNames , freq_array , reffreq_array,
                                    array_dict_loc, x_new=None, rigoureous=True,  order_axis=0): 
    """ 
    Interpolate value from  reference frequency : 
 
    :param stationNames:  list of stations of survey .
    :type stationNames: list
     
    :param freq_array:  array of frequencies 
    :type freq_array: (ndarray,1)
     
    :param reffreq_array: reference frequence for clean Data .
    :type reffreq_array: float 
     
    :param array_dict_loc: location of each value according each station names  
                            keys are stationsNmes and value of array
                            at that station , eg -S00 , valueof RhO  
    :type array_dict_loc: dict 
    """
    
    def _interpolate_array(array_to, kind='linear', fill_value='extrapolate' ):
        """ 
        Interpolate value to according the list of frequencies. 
        
        :param array_to: (ndarray,1), array to interpolate
        :type array_to: array_like
        
        :param fill_value: kind of extrapolation
        :type fill_value: str 
        """ 

        Y_intp_value =[]
        for ss, rowline in enumerate(array_to):
            func_interp = spi.interp1d(x_old, rowline, kind='linear', fill_value=fill_value)
            y_new=func_interp(x_new)

            Y_intp_value.append(y_new)
        new_Y_array =func.concat_array_from_list(list_of_array=Y_intp_value, concat_axis=order_axis)
        
        return new_Y_array
    
    if x_new is None : x_new=freq_array
    x_old =freq_array[:get_reffreq_index(freq_array=freq_array,
                                                    reffreq_value=reffreq_array)+1] # add the value of rffreq_freqq
    if rigoureous is True : 
        x_new = np.linspace(freq_array[0], freq_array[-1], x_old.size)

    temp_list =[values [:get_reffreq_index(freq_array=freq_array,
                                                     reffreq_value=reffreq_array)+1] for  keys , values in array_dict_loc.items() \
                                                       for  stn in stationNames  if keys ==stn  ]
    value_to_intp =func.concat_array_from_list(list_of_array=temp_list, concat_axis=order_axis) 
    return value_to_intp , _interpolate_array(array_to=value_to_intp)
       

def find_reference_frequency(freq_array =None, reffreq_value =None , sharp =False, etching=True): 
    """
    Method to find and interpolate reference value if it is not present on the frequency range. 

    :param freq_array: array_like frequency range 
    :type freq_array: array_like
    
    :param reffreq_value: reference frequency value
    :type reffreq_value: float or int 
    
    :param sharp:  if set to True , it forces the program to find mainly 
                    a value closest inside the  frequency range.
    :type sharp: bool  

    :param etching: bool , if set to True , it will print in your stdout.
    :type etching: bool 
    
    :returns: reference frequency 
    :rtype: float 
            
    """

    if freq_array is None or reffreq_value is None : 
        raise ValueError("None value can not be computed. check your frequency array or reference value.")
        
    if freq_array.dtype not in ['float', 'int']: 
        try : freq_array=np.array([float(ff) for ff in freq_array])
        except :raise TypeError('Frequency data must be on (ndarray,1) of float value not str.')
    if type(reffreq_value) is not int or type(reffreq_value) is not float: 
        try : reffreq_value= np.int(reffreq_value)
        except: raise CSex.pyCSAMTError_frequency("Reference frequency must be either float value or integer.")
        
    if freq_array.min() > reffreq_value > freq_array.max():
        warnings.warn('Reference frequency is out of frequency range. see more infos about reference '\
                                                  ' frequency.|{0}|'.format(infos.notion.reference_frequency))
        raise CSex.pyCSAMTError_frequency ("Input reference frequency is out the frequency range. "\
                                          "Please put value between the frequency range. "\
                                        "Frequency range is [{0} Hz to {1} Hz].".format(freq_array.min(), freq_array.max()))
    
    def force_interpolation (value_to_steep= None): 
        """ 
        Method to force reference value interpolated to find a value in frequency range close to . 

        :param value_to_steep:  reference frequency value to be forced .
        :type value_to_steep: float 
        
        :returns: closet value of interpolated frequency 
        :rtype: float
        """ 
        #find the value close to reference array.
        
        if freq_array[0] > freq_array[-1]: # check whether frequencies are range to higher frequency to lower :  
            temp_freq_array= freq_array[::-1] # sorted to lower to Highest 
        else : temp_freq_array = freq_array
        
        if value_to_steep is not None : 
            for xi in temp_freq_array : 
                if xi == value_to_steep : 
                    index_refreq= np.where(temp_freq_array==xi)[0] #find the index location on the normalization freq 
                    break 
                elif xi > value_to_steep : 
                    # check which frequency is closet one : 
                    index_temp= np.where(temp_freq_array==xi)[0] # find the upper location and take the lower value 
                                                       #assume to be the -1nieme element of the range.
                    # print(xi, index_temp)
                    dx1 = np.abs(temp_freq_array[index_temp-1]- xi )
                    dx2 = np.abs(temp_freq_array[index_temp] -xi)
                    if dx1 < dx2 : index_refreq =index_temp-1 
                    else : index_refreq=index_temp
                    break
        #new_value find  . and 
            new_val = temp_freq_array[index_refreq]
        return freq_array [np.where(freq_array ==new_val)]
        
        # return freq_array[index_refreq]
    
    interp_func =spi.interp1d(x=np.log10(freq_array), y=freq_array)
    new_reference_value = interp_func(np.log10(reffreq_value))

    if sharp ==True :
        if etching:
            print('---> Input reference frequency <{0}> Hz has been interpolated to < {1} > Hz.'.\
                   format(np.float(reffreq_value), np.around(float(new_reference_value ),2)))
        return force_interpolation(value_to_steep=new_reference_value  )
    elif sharp ==False :return new_reference_value 



     
                                                              
def get_data_from_reference_frequency(array_loc, freq_array, reffreq_value):
    """
    Function to get reference frequency  without call especially stations array.
    The function is profitable but . It's less expensive However if something wrong happened
    by using the first step to get a reference array , it will try the traditionnally function 
    to get it. If none value is found , an Error will 
    occurs. 

    :param array_loc: assume to be a dictionnary of stations_data_values. 
    :type  array_loc: dict
    
    :param freq_array:  frequency array 
    :type  freq_array: array_like
    
    :param reffreq_value:  reffrence value, If the reference value is not in frequency array ,
                            function will force to interpolate value and find the correlative array.
                           
    :type reffreq_value: float or int 
            
    :returns:  an array of reference value at specific index 
    :rtype: array_like 
            
    :Example: 
        
        >>> get_data_from_reference_frequency(array_loc=rho,freq_array=freq_array,
        ...                                      reffreq_value=1023.)
        ... Input reference frequency has been interpolated to < 1024.0 > Hz 
    """
    
    if type(reffreq_value) is str : 
        try: reffreq_value=np.float(reffreq_value)
        except:raise CSex.pyCSAMTError_frequency("Reference value must be float or int value, not str!")
    #find the size of array loc : seem to be the size of stationsNames .
    stn_size =np.array(list(array_loc.keys())).size
    # # we sorted data for ensurity
    reffreq_array= np.array([values[np.where(freq_array==reffreq_value)[0]]
                             for keys, values in sorted(array_loc.items())])

    if reffreq_value not in freq_array : 
        reffreq_array= get_data_from_reference_frequency(array_loc=array_loc,
                                                         freq_array=freq_array,
                                                         reffreq_value=find_reference_frequency(
                                                                      freq_array =freq_array, 
                                                                      reffreq_value =reffreq_value ,
                                                                      sharp =True, 
                                                                      etching=True))
    if reffreq_array.size == stn_size :
        return reffreq_array.reshape(reffreq_array.shape[0],)
    else :

        from pycsamt.processing.callffunc import get_array_from_reffreq  as gfreq
        reffreq_array=gfreq(array_loc=array_loc , freq_array=freq_array, 
                            reffreq_value=reffreq_value,
                            stnNames=sorted(array_loc.keys()))
        if reffreq_array is None :
            raise CSex.pyCSAMTError_frequency('Something wrong happened during reference array computing.'\
                                             ' Please check the your inputs data size, and your station size.')
        return reffreq_array

def perforce_reference_freq(dataset, frequency_array=None ):
    """ 
    Function to get automatically the reference frequency. If user doesnt provide the value , 
    function will find automatically value . 
 
    :param data: array of avg DATA,  ndim>1 
    :type data: array_like 
    
    :param frequency_array: array of frequency
    :type frequency_array: array_like   
    
    :returns: reffreq_value float , reference frequency value 
    :rtype: float
    
    :returns: uncover_index,  index of reference value on frequency array 
    :rtype: int 
     
    :returns: nan_ratio , the ratio or the prevalence of nan in the data_set 
    :rtype: float 
    """
    import copy
    new_data_array =copy.deepcopy(dataset)
    if frequency_array is None :
        frequency_array,*_=np.unique(new_data_array[:,1], return_counts=True) #find according to AVGData disposal .
        
    # ifall data are cleaned the take the highst freqeuncy 
    if np.isnan(new_data_array).sum()/new_data_array.size ==0: 
        if frequency_array[0] < frequency_array[-1] :
            reffreq_value = frequency_array[-1] 
        else : reffreq_value = frequency_array[0]
        print('--> Reference frequency is estimated to <{0}> Hz'.format(reffreq_value ))
        
        return  reffreq_value , frequency_array.size -1,0.
    
    elif np.isnan(new_data_array).sum()/new_data_array.size !=0 : 
        nan_ratio = np.isnan(new_data_array).sum()/new_data_array.size 
        new_data_array=np.nan_to_num(new_data_array, nan = -99999)
        # uncover_index = np.where(new_data_array==-99999)   
        reffreq_value = frequency_array[np.where(new_data_array==-99999)[0][0]]
        print('--> Reference frequency is estimated to <{0}> Hz'.format(reffreq_value))
        return reffreq_value, np.where(new_data_array==-99999)[0][0],  nan_ratio 

def hanning (dipole_length=50., number_of_points=7, large_band=False):
    """ 
    Function to compute hanning window .
    
    .. seealso::Torres-Verdin and Bostick, 1992, Principles of spatial surface electric
             field filtering in magnetotellurics: electromagnetic array profiling (EMAP),
             Geophysics, v57, p603-622.
        
    :param dipole_length: the length of dipole , xk is centered between dipole 
    :type dipole_length: float
    
    :param number_of_points: number of filter points 
    :type number_of_points: int  
    
    :returns: windowed hanning 
    :rtype: array_like 
    
    """
    
    if type(dipole_length) !=float or type(number_of_points) !=int: 
        
        try : 
            dipole_length = np.float(dipole_length)
            number_of_points = np.int(np.floor(number_of_points))
        except :
            raise CSex.pyCSAMTError_processing('Dipole length and number of'\
                                           ' points must be a float number.')
        
    window_width = number_of_points * dipole_length
    xk , HAN = window_width / 2 , []                           # center point xk 
    X= np.linspace(-xk, xk, number_of_points)
    # X = np.arange(  -xk , xk, dipole_length)
    # xpos= np.arange(number_of_points+1)*xk 
    
    if large_band == True : return np.array([(1/window_width * \
                                              (1+ np.cos(2* np.pi * xx / window_width))) for xx in X ])

    for xx in X : 
        if xx <= 0: hannx= 1/window_width * (1+ np.cos(2* np.pi * xx / window_width))
        else :hannx = 0 
        HAN.append(hannx)
    # import matplotlib.pyplot as plt 
    # plt.plot(X,np.array(HAN) )
    # plt.show()

    return np.array(HAN)

 
def weight_beta (dipole_length =50. , number_of_points = 7, window_width=None): 
    """
    WeightBeta function is  weight Hanning window . if window width is not provide  , function 
    will compute the width of window. 
    
    .. seealso:: 
            Torres-Verdin and Bostick, 1992, Principles of spatial surface electric field filtering in
            magnetotellurics: electromagnetic array profiling (EMAP), Geophysics, v57, p603-622.
            ...
            
    .. note::SUM(Betaj (j=1..M)=1 .
              
    :param dipole_length: length of dipole in meter (m) 
    :type dipole_length: float
    
    :param number_of_points: number of station points to filter
    :type  number_of_points: int 
    
    :param window_width: the width of window filter 
    :type window_width: float 
      
    :returns: beta_array at each station 
    :rtype: array_like 
    """
    _logger.info( 'Computing weight Hanning window ')
    # beta_infx = - dipole_length/2 
    # beta_posfx = dipole_length /2 
    if window_width is None : window_width = number_of_points *  dipole_length
    if window_width is not None : 
        try : number_of_points = np.int(window_width / dipole_length ) 
        except : 
            warnings.warn('Please see more info about coeff.Beta :<{0}>'.format(infos.notion.weighted_Beta))

            raise CSex.pyCSAMTError_processing('Window width type must be float number.')
    
    xk = window_width / 2 
    num_pk =np.arange(dipole_length/2, window_width , dipole_length)
    beta_range= []
    for xp in num_pk : 
        if np.abs(xp - xk) <= window_width /2 : 
            beta_xp = ((1/window_width) * ((xp + dipole_length/2 - xk) +(window_width/ 2* np.pi)*\
                       (np.sin(2* np.pi * (xp + dipole_length/2 -xk )/window_width))))-\
                ((1/window_width)* ((xp - dipole_length/2 - xk) +(window_width/ 2* np.pi)*\
                 (np.sin(2* np.pi * (xp - dipole_length/2 -xk)/window_width))))
                    
        elif np.abs(xp - xk) > window_width/2 : beta_xp = 0.
        
        beta_range.append(beta_xp) 
    
    return np.array(beta_range)
    # for xj in enra
    
@deprecated('Function deprecated') 
def compute_adaptative_moving_average ( z_array =None  , weighted_window=None ,
                  dipole_length = None , number_of_points =None ):
    """
    .. note:: 
        see `compute_AMA` !
    """
     
    if z_array is None : raise CSex.pyCSAMTError_Z('NoneType Impedance can not be computed.')
    if weighted_window is None : 
        if dipole_length is None or number_of_points is None :
            raise CSex.pyCSAMTError_inputarguments('Please check '
                'your Inputs values ! NoneType values of dipole length or number_of_stations'
                ' can not be computed.!')
        try : 
            dipole_length , number_of_points  = np.float(dipole_length) , np.int(number_of_points)
        except : raise CSex.pyCSAMTError_float('Dipole length, number of points must be a float, and '\
                                               ' int  number respectively.!')

        weighted_window =weight_beta(dipole_length=dipole_length, 
                                      number_of_points=number_of_points)

    xc_pk , cFLAM , cut_off = np.int(np.floor(weighted_window.size / 2)), [], -1 

    for kk , zz in enumerate(z_array): 
        if kk < xc_pk : 
            cut_off += 1 
            zz_new = weighted_window[xc_pk - cut_off : weighted_window.size] *\
            z_array[kk - cut_off: weighted_window[xc_pk-cut_off : weighted_window.size].size]
            cFLAM.append(zz_new.sum())
        elif kk == xc_pk or kk <= z_array.size - weighted_window[xc_pk:].size: 
            zz_new, cut_off= weighted_window * z_array[kk- xc_pk : kk + weighted_window[xc_pk :\
                                                                weighted_window.size].size], 0
            cFLAM.append(zz_new.sum())
        elif kk > z_array.size - weighted_window[xc_pk:].size: 
            cut_off -= 1
            zz_new = z_array[kk-xc_pk :] * weighted_window[:(weighted_window.size +cut_off)]
            cFLAM.append(zz_new.sum())
            
    return np.array(cFLAM)
            
def hanning_xk (dipole_length =50. , number_of_points =7):
    """
    compute _hanning window on a wtdth of number of point : 
        integrate value on all the window_bandwidth discrete and continue.
        if value is greater than Hald of the width  value == 0 . 
       
    :param dipole_length: length of dipole 
    :type dipole_length: float 
    
    :param number_of_points:  value of points or survey stations . 
    :type number_of_points: int 
    
    :returns: han_xk , continue value on half bandwidth x0-- xk (center point) 
    :rtype: array_like 
    
    :returns: windowed hanning, discrete _value ,SUM(han(x0, xk))
    :rtype: array_like 
    
    """
    
    window_width = number_of_points *  dipole_length 
    xk , hannx= window_width / 2 , []
    xpk_num =np.arange(dipole_length/2, window_width, dipole_length)

    han_xk = np.array((1/ window_width * ( xk  + (window_width/2*np.pi ) * np.sin(2 * np.pi * (xk) /window_width))) - \
                (1/ window_width * ( xpk_num[0]  + (window_width/2*np.pi ) * np.sin(2 * np.pi * (xpk_num[0]) /window_width))))
    
    # compute Hanning discrete : 
    for  ii, xx in enumerate(xpk_num) : 
        if np.abs(xx) <= window_width /2 :
            han_xx = 1/ window_width * ( (xx + dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xx + dipole_length/2) /window_width))) - \
                1/ window_width * ( (xx - dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xx - dipole_length/2) /window_width)))
        else : han_xx= 0 
        
        hannx.append(han_xx)

    
    return han_xk , np.array(hannx)
    
def hanning_x (x_point_value, dipole_length =50.,  number_of_points=7, 
               bandwidth_values=False , on_half=True):
    """
    Function to compute point on window width .  Use discrete computing . Function show the value
    at center point of window assume that the point is center locate on the window width . 
    It intergrates  value between dipole length. User can use see_extraband to see the values 
    on the total bandwith. If half is False the value of greater than center point will be
    computed and not be 0 as the normal definition of Hanning window filter. 
    
    :param x_point_value:  value  to intergrate.
    :type x_point_value: float 
    
    :param dipole_length: length of dipole on survey
    :type dipole_length: float 
    
    :param number_of_point: survey point or number point to apply.
    :type number_of_point: int 
    
    :param bandwidth_values: see all value on the bandwith , value greater than x_center
                            point will be computed .
    :type bandwidth_values:bool 
                            
    :param on_half:  value on the bandwith; value greater that x_center point = 0.
    :type on_half: bool 

    :returns: hannx  integrated X_point_value or  array of window bandwidth .
    :rtype: array_like 
    """
    window_width = number_of_points * dipole_length 
    
    xx_points_array = np.arange(dipole_length/2 , window_width , dipole_length )
    
    def see_extraband( apply_on_half =False): 
        """ 
        Function to see data of the large bande integrated hanning. if apply_on_half 
        is set to True < it will show the data on band <= half windowith i.e. thin 
        center point xk.
        
        :param apply_on_half: half band of hanning filter.
        :type apply_on_half: bool  

        :returns: windowed hanning band
        :rtype: array_like(ndarray,1) 
        """
        hv=[]
        
        if not apply_on_half:
            for xv  in  xx_points_array:
                hann_xv= np.array( 1/ window_width * ( (xv + dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xv+ dipole_length/2) /window_width))) - \
                        1/ window_width * ( (xv - dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xv - dipole_length/2) /window_width))))
                hv.append(hann_xv)
        elif apply_on_half : 
            for xv  in  xx_points_array:
                if xv  <= window_width/2: 
                    hann_xv= np.array( 1/ window_width * ( (xv + dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xv+ dipole_length/2) /window_width))) - \
                        1/ window_width * ( (xv - dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (xv - dipole_length/2) /window_width))))
                else :hann_xv=0 
                hv.append(hann_xv)
                
        return np.array(hv)
        
    if (x_point_value - dipole_length/2 < 0 or x_point_value +dipole_length/2 > window_width):
        raise CSex.pyCSAMTError_processing('Hanning x point value is outside the window.')
    
    if bandwidth_values is True : 
        if on_half :return see_extraband(apply_on_half=True)
        if not on_half :return  see_extraband(apply_on_half=False)
    else :
        if x_point_value <= window_width/2:
            hann_xp= np.array( 1/ window_width * ( (x_point_value + dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (x_point_value + dipole_length/2) /window_width))) - \
                        1/ window_width * ( (x_point_value - dipole_length/2)  + (window_width/2*np.pi  * np.sin(2 * np.pi * (x_point_value - dipole_length/2) /window_width))))
            
        else : hann_xp = 0
        
    return  hann_xp

def wbetaX2 (Xpos, dipole_length=50., number_of_points=5) :
    """
    weight Beta is  computed following the paper of 
    Torres-verdfn, C., and F. X. Bostick, 1992, Principles of spatial surface electric field 
    filtering in magnetotellurics - Electromagnetic array profiling ( EMAP )- Geophysics, 57(4), 25–34. 
        
    :param Xpos: reference position on the field 
    :type Xpos: str 
    
    :param dipole_length: length of dipole measurement
    :type dipole_length: float 
    
    :param number_of_points:  point to stand filters , window width 
    :type number_of_points: int 
    
    """
    
    window_width = dipole_length * number_of_points 
    xk = window_width/2
    

    # if np.abs(xp - xk) <= window_width /2 : 
    beta_xpos = ((1/window_width) * ((Xpos + dipole_length/2 - xk) +(window_width/ 2* np.pi)*\
               (np.sin(2* np.pi * (Xpos + dipole_length/2 -xk )/window_width))))-\
        ((1/window_width)* ((Xpos - dipole_length/2 - xk) +(window_width/ 2* np.pi)*\
         (np.sin(2* np.pi * (Xpos - dipole_length/2 -xk)/window_width))))
            
        # elif np.abs(xp - xk) > window_width/2 : beta_xpos = 0.
        
        # beta_range.append(beta_xpos) 
    return np.array([Xpos-dipole_length/2, Xpos + dipole_length/2]) , beta_xpos 

def compute_sigmas_e_h_and_sigma_rho(pc_emag , pc_hmag, pc_app_rho, app_rho, emag, hmag):
    """
    function to compute Standard Deviation for E-field (sigma_e), 
    Standard Deviation for H-Field (sigma_h) , 
    & Standard Deviation for Component RHO (sigma_rho)

    Parameters
    -----------
        * pc_emag : float  
                Statistical variation of magnitude values from averaged data blocks.
                Standard Deviation/Average Emag (%)
        * pc_hmag :float 
                Statistical variation of magnitude values from averaged data blocks.
                Standard Deviation / Average Hmag (%)
        * pc_app_rho: float 
                Statistical variation of magnitude values from averaged data blocks.
                Standard Deviation / Average Rho (%)
                 
        * app_rho :float 
                resistivity calculated from averaged component (ohm.m)
        * Emag : float
                average E - field magnitude(microVolt/Km *amp )
        * Hmag : float 
                average H - field magnitude(pTesta/amp) or (milliGammas/Amp)
        
    Returns
    --------
        sigma_rho : float 
             srhoC (Standard Deviation for Component RHO)
        c_var_Rho : float 
            C-varrhoC(  Coefficient of Variation for Component RHO)
    """
    # Standard Deviation for E-field (se) in (%) 1micoV/m*amp= 1mV/Km*amp
    emag /=1e3 # convert microvolt to millivolt 
    # sigma_e, sigma_h, sigma_rho   = emag *(pc_emag/100), hmag * (pc_hmag /100) ,\
    #     pc_app_rho/100 * app_rho
    
    sigma_e , sigma_h, sigma_rho   =  pc_emag/100, pc_hmag /100 , pc_app_rho/100

   
    return  sigma_e , sigma_h, sigma_rho 
    
def compute_weight_factor_betaj(Xpos , dipole_length=50. , number_of_points =5. , 
                                window_width=None):
    """
    weight Beta is  computed following the paper of Torres-verdfn, C., and F. X. Bostick, 1992,
     Principles of spatial surface electric field filtering in magnetotellurics - Electromagnetic 
    array profiling ( EMAP )- Geophysics, 57(4), 25–34. 
        
    :param Xpos: reference position on the field 
    :type Xpos: str 
    
    :param dipole_length: length of dipole measurement
    :type dipole_length: float 
    
    :param number_of_points:  point to stand filters , windowed width 
    :type number_of_points: int 
    
    """
    #figureout the center ot dipole_lenght  and  compute the window width 
    if window_width is None : 
        window_width = dipole_length * number_of_points  #+  dipole_length # for extrema 
    
    xk = window_width/2
    
    # declare limit of integrating hanning 
    
    xsup_lim = Xpos + dipole_length/2 
    xinf_lim = Xpos - dipole_length /2 
    
    X =np.arange(0, window_width, dipole_length) 

    betapos = (1/ window_width) * ((xsup_lim - xk + (window_width /(2 * np.pi)) *\
                                 np.sin( 2* np.pi *(xsup_lim - xk)/ window_width )) - (
                                xinf_lim - xk + window_width /(2 * np.pi) *\
                                 np.sin( 2* np.pi *(xinf_lim - xk)/ window_width ))) 
            
    return betapos, window_width, X 

def compute_FLMA ( z_array =None  , weighted_window=None ,
                  dipole_length = 50. , number_of_points =5., **kwargs ):
    """
    Use a fixed-length-moving-average filter to estimate average apparent
    resistivities at a single static-correction-reference frequency
    
    :param z_array:  array of uncorrected impedance values.
    :type z_array: arry_like, complex
    
    :param dipole_length: length of dipole on survey
    :type dipole_length: float 
    
    :param number_of_points: survey point or number point to apply.
    :type number_of_points: int 
    
    :param weighted_window: bandwidth of hanning window filter lengths,
                           skin_depth arrays at each station.
    :type weighted_window: array_like 
 
    :param app_rho: array of apparent resistivities in ohm.m , if not provided 
                    will use impedance zrho to compute app_rho with reference_frequency 
    :type app_rho: array_like 
    
    :returns: windowed hanning , filtered window.
    :rtype: array_like 
    
    :Example:
        
        >>> from pycsamt.ff.core.processing import zcalculator as Zcc
        >>> z1= np.array([2.46073791 +3.00162006j ,
        ...             9.74193019 +1.82209497j,
        ...             15.68879141 + 12.91164264j ,
        ...             5.84384925 +3.6899018j, 
        ...             2.4430065  +0.57175607j])
        >>> flma= compute_FLMA(z_array=z1, dipole_length=50. ,
                               number_of_points=4)
        >>> print(flma)
    """
    #reference_freq= kwargs.pop('reference_freq', 8192.)
    
    if z_array is None : 
        raise CSex.pyCSAMTError_Z('NoneType Impedance can not be computed.')
    if weighted_window is None : 
        if dipole_length is None or number_of_points is None :
            raise CSex.pyCSAMTError_inputarguments('Please check '
                'your Inputs values ! NoneType values of dipole length '
                ' or number_of_stations can not be computed.!')
        try : 
            dipole_length = np.float(dipole_length) 
            number_of_points= np.int(number_of_points)
            
        except :
            raise CSex.pyCSAMTError_float(
                'Dipole length, number of points must be a float, and '
                    ' int  number respectively.!')

        weighted_window = np.arange(0, dipole_length * number_of_points , 
                                    dipole_length)
       
    # compute weight factor Bj 
    wk = np.array([compute_weight_factor_betaj(Xpos , dipole_length=dipole_length , 
                                number_of_points =number_of_points)[0] 
          for Xpos in weighted_window])
    
    #seek the center point  index xc_pk  

    xc_pk =  np.int(wk.size /2)   # weight array size is odd 
       
    
    cFLMA , cut_off = [], -1  # take a minimum 

    import copy 
    tem_wk = copy.deepcopy(wk)
    
    for kk , zz in enumerate(z_array): 
        if kk < xc_pk : 
            cut_off += 1  # cut off to shrink one index  
            tem_wk [:xc_pk -cut_off] = 0. # where wk 0 outside of the Hanning window
            tem_z= z_array[:kk + xc_pk +1] # keep main array 
            # buid array with number of 0 outside the window 
            add0 = len(wk)- len(tem_z)  
            # build array as the same dim with hanning window centered a k position
            tem_z = np.concatenate((np.repeat(0., add0), tem_z))
            zz_new= tem_z * tem_wk # compute value 
            cFLMA.append(zz_new.sum())
            
            tem_wk =copy.deepcopy(wk) # copy once again for next loop if exist 
            
        elif xc_pk <= kk:
            if len(wk) %2 !=0: # if window weight len is odd 
                if  kk < z_array.size - xc_pk:
                    zz_new= z_array[kk - xc_pk: kk + xc_pk +1  ] * wk # when center point is even 
                elif  kk >= z_array.size - xc_pk:
                    tem_z = z_array[kk-xc_pk:] 
                    add0 = len(wk) - len(tem_z)  
                    tem_wk = np.concatenate((tem_z, np.repeat(0.,add0)))
                    zz_new =tem_wk * wk

            elif len(wk) %2 ==0:# when window length is even.
                if kk <= z_array.size - xc_pk:
                    zz_new= z_array[kk - xc_pk  : kk + xc_pk] * wk
      
                elif  kk > z_array.size - xc_pk:
                    tem_z = z_array[kk-xc_pk:] 
                    add0 = len(wk) - len(tem_z)  
                    tem_wk = np.concatenate((tem_z, np.repeat(0.,add0)))
                    zz_new =tem_wk * wk    
                    
            cFLMA.append(zz_new.sum())    
            
    print('** {0:<27} {1} {2} m.'.format("Filter's width",
                                '=', int(number_of_points)* dipole_length))   
             
    return np.array(cFLMA)    

def compute_TMA (data_array=None, number_of_TMApoints=5. ):
    """
    function to compute a trimmed-moving-average filter 
    to estimate average apparent resistivities.

    :param data_array:  content of value to be trimmed 
    :type data_array: array_like(ndarray,1) 
    
    :param number_of_TMA points:  number of filter points .
    :type number_of_TMA points: int 
    
    :returns:  value corrected with TMA  
    :rtype: array_like (ndarray, 1)
    
    :Example:
        
        >>> from pycsamt.ff.core.processing import zcalculator as Zcc
        >>> z2 = np.array([2.46073791 , 3.00162006 ,
        ...     9.74193019 ])# 1.82209497,
        ...     # 15.68879141 ,  12.91164264 ,
        ...     # 5.84384925 , 3.6899018, 
        ...     # 2.4430065  ,0.57175607])
        >>> tma= Zcc.compute_TMA(data_array=z2,
        ...                     number_of_TMApoints= 7)
        ... print(tma)

    """
    
    if data_array is None :
        raise CSex.pyCSAMTError_processing(
            'NoneType arguments can not be computed!.')
    
    if data_array.dtype not in ['float', 'int']: 
        try :data_array=np.array([float(dd) for dd in data_array])
        except : 
            raise CSex.pyCSAMTError_float(
                'Data must be on array_like float number.!')
    if type(number_of_TMApoints) is not int :
        try : 
            number_of_TMApoints=np.int(number_of_TMApoints)
        except : 
            raise CSex.pyCSAMTError_parameter_number(
                'TMA filter point must be integer.')
    
    roll_TMA =[]
    # let compute TMA for one points, 1 poinst will return the data array
    # Ascertqin number of TMA point and Data _array
    
    if number_of_TMApoints > len(data_array) : 
        warnings.warn("Number of TMA points = {}  is too large !".
                      format(number_of_TMApoints))
        number_of_TMApoints =5 # reset temporray to 5.
        
    if len(data_array) < 5: #find the TMA point accessible for computation
        # window width 
        number_of_TMApoints = len(data_array)
        
        if len(data_array) <= 2:
            
            warnings.warn(" Number of sites explored are too short !")
            print("--> Investigation sites too short !")
            
            print('** {0:<27} {1} {2} pts.'.format('Filter width  ',
                                '=', int(number_of_TMApoints))) 
            
            if len(data_array) ==1 :  return data_array 
            if len(data_array) ==2 : # take the mean of both value 
                return np.repeat(data_array.mean(), len(data_array))
        
        if len(data_array)==3: # loop the window and take mean value at each time
            print("--> ! Less number of TMA points  for higher frequencies mitigation.!")
                  
            for kk in range(len(data_array)): # add add overall mean to the last value 
                if kk ==len(data_array)-1:
                    tma = np.array([data_array.mean(),
                                    data_array[-1]]).mean()
                    roll_TMA.append(float(tma)) 
                else : 
                    roll_TMA.append(data_array[kk:kk+2].mean())
            
            print('** {0:<27} {1} {2} pts.'.format('Filter width  ',
                                '=', int(number_of_TMApoints)))     
            return np.array(roll_TMA)
        
     
    if len(data_array) >= 5 and number_of_TMApoints < 5: 

        if number_of_TMApoints <= 3 :
            warnings.warn("Number of points provided to compute TMA are not enough "
                          "for higher frequencies reduction.TMA points less than three (3)  "
                          " are not consistent to correct data."
                          )
            print("-->  Number of points are not enough for higher frequencies attenuation."
                  "For consistency, provide TMA points more than three.")
        
        if len(data_array)== 5 : mesf ="equal to"
        else : mesf = "more than"
        
        warnings.warn("Number of sites explored is estimated "
                      "to {}, then  number of TMA points is resetting "
                      "to 5 as default value.".format(len(data_array)))
        
        print("--> Number of explored sites is {0}  five(5).".format(mesf))
        print("--> TMA points was resseting  to 5.")
        
        number_of_TMApoints=5
        
    
    xk_cp  = np.int(np.trunc(number_of_TMApoints/2)) # centered point
    
    cut_off=-1          #starting point 
    for kk, value in enumerate(data_array): 
        if kk < xk_cp : 
            cut_off +=1 
            tma= data_array[kk-cut_off: kk+ xk_cp+1]
            
        elif xk_cp <= kk <= data_array.size - xk_cp:
            tma = data_array[kk- xk_cp:kk + xk_cp +1] # for Python index 
        
        elif kk > data_array.size - xk_cp:
            tma = data_array[kk-xk_cp:]

        get_TMA_value= np.delete(tma, 
                       np.array(tma.argmin(), tma.argmax())
                        )
        #tma = np.delete(tma, np.argmax(rho_tem), axis = 0)
        roll_TMA.append(get_TMA_value.mean())
        
    print('** {0:<27} {1} {2} pts.'.format('Filter width  ',
                                '=', int(number_of_TMApoints)))     
    return np.array(roll_TMA)   

@deprecated('Function too  expensive, redirected to `compute_TMA` more faster.')    
@redirect_cls_or_func(compute_TMA, "short and faster than the func below")    
def compute_trimming_moving_average (data_array=None, number_of_TMApoints=None ):
    """
    function to compute a trimmed-moving-average filter
    to estimate average apparent resistivities.

    :param data_array:  content of value to be trimmed 
    :type data_array: array_like(ndarray,1) 
    
    :param number_of_TMA points:  number of filter points .
    :type number_of_TMA points: int 
    
    :returns:  value corrected with TMA  
    :rtype: array_like (ndarray, 1)
          
    :Example:
        
        >>> from pycsamt.ff.core.processing import zcalculator as Zcc
        >>> z2 = np.array([2.46073791 , 3.00162006 ,
        ...     9.74193019 ])# 1.82209497,
        ...     # 15.68879141 ,  12.91164264 ,
        ...     # 5.84384925 , 3.6899018, 
        ...     # 2.4430065  ,0.57175607])
        >>>  tma= Zcc.compute_trimming_moving_average(data_array=z2,
        ...                                          number_of_TMApoints= 4)
        >>> print(tma)
        >>> print(z2.shape)
        >>> print(tma.shape)
    """
    
    if data_array is None or number_of_TMApoints is None :
        raise CSex.pyCSAMTError_processing('NoneType arguments can not be computed!.')
        
    if data_array.dtype not in ['float', 'int']: 
        try :data_array=np.array([float(dd) for dd in data_array])
        except : raise CSex.pyCSAMTError_AvgData('Data must be on array_like float number.!')
    if type(number_of_TMApoints) is not int :
        try : number_of_TMApoints=np.int(number_of_TMApoints)
        except : raise CSex.pyCSAMTError_parameter_number('TMA filter point must be integer.')
    
    roll_TMA,window  =[], np.int(np.trunc(number_of_TMApoints/2)),

    for ii, value in enumerate(data_array): 
        if ii < window:
            get_TMA_value =np.delete(data_array[ii:number_of_TMApoints+ii], 
                                     np.array([data_array[ii:number_of_TMApoints+ii].argmin(), 
                                     data_array[ii:number_of_TMApoints+ii].argmax()])).mean()
            roll_TMA.append(get_TMA_value)
            
        elif ii == window or ii < data_array.size - window :
            if ii == data_array.size - (window+1) :
                get_TMA_value =np.delete(data_array[ii-window:],
                                         np.array([data_array[ii-window:].argmin(),
                                                                                                               data_array[ii-window:].argmax()])).mean() 
            else :get_TMA_value =np.delete(data_array[ii-window:ii+window+1],
                                           np.array([data_array[ii-window:ii+window+1].argmin(), 
                                           data_array[ii-window:ii+window+1].argmax()])).mean()
            roll_TMA.append(get_TMA_value)
            
        elif ii >=data_array.size - window:
            if value == data_array[-1]:
                get_TMA_value =np.delete(data_array[ii-number_of_TMApoints-1:],
                                         np.array([data_array[ii-number_of_TMApoints-1:].argmin(), 
                                         data_array[ii-number_of_TMApoints-1:].argmax()])).mean()
                
            else :get_TMA_value =np.delete(data_array[ii-number_of_TMApoints:],
                                           np.array([data_array[ii-number_of_TMApoints+1:ii+1].argmin(), 
                                           data_array[ii-number_of_TMApoints+1:ii+1].argmax()])).mean()  
            roll_TMA.append(get_TMA_value) 
            
    return np.array(roll_TMA)  
           
def compute_AMA(reference_freq=None, z_array =None, 
                 number_of_skin_depth=1., dipole_length = 50., 
                  **kwargs  ):
    """
    Use an adaptive-moving-average filter to estimate average apparent
    resistivities at a single static-correction-reference frequency. Adaptive
    filtering is based on ideas presented inTorres-Verdin and Bostick, 1992,
    Principles of spatial surface electric field filtering in magnetotellurics
    -electromagnetic array profiling (EMAP), Geophysics, v57, p603-622.
    
    :param reference_frequency: frequency with clean data in Hertz 
    :type reference_frequency: float 
    
    :param z_array:  array of uncorrected impedance values.
    :type z_array: arry_like, complex
    
    :param dipole_length: length of dipole on survey
    :type dipole_length: float 
    
    :param number_of_skin_depth: arbitrary real constant, can be changed 
                to any value between 1 and 10 skin depths, by experiency, 
                value shoulb be 1<= c <= 4
    :type number_of_skin_depth: int , default is 1.
    
    :param weighted_window: bandwidth of hanning window filter lengths,
                           skin_depth arrays at each station.
    :type weighted_window: array_like 
 
    :param app_rho: array of apparent resistivities in ohm.m , if not provided 
                    will use impedance zrho to compute app_rho with reference_frequency 
    :type app_rho: array_like 
    
    :returns: windowed hanning , filtered window.
    :rtype: array_like 
    
    .. note:: If resistivities array is provided, provided also uncorrected phase
            array  to compute `impedance array`.
    
    :Example:
        
        >>> from pycsamt.ff.core.processing import zcalculator as Zcc
        >>> z1= np.array([2.46073791 +3.00162006j ,
        ...     9.74193019 +1.82209497j,
        ...     15.68879141 + 12.91164264j ,
        ...     5.84384925 +3.6899018j, 
        ...     2.4430065  +0.57175607j])
        ...
        >>> reference_frequency = 8192.
        >>> znew= compute_AMA(reference_frequency=8192.,
        ...                   number_of_skin_depth= 1., dipole_length =50.,
        ...                   z_array=z1)
        ... print(znew)
    
    """
    
    if reference_freq is None : 
        raise CSex.pyCSAMTError_frequency('Please add your reference frequency!')
        
    mu0 = 4* np.pi * 1e-7 
    omega_reffreq  = 2* np.pi *  reference_freq  
    
    resistivities = kwargs.pop('app_rho', None)
    phase =kwargs.pop('phase', None)
    adaptative_value =kwargs.pop('scale', 10.)

    if resistivities is None : 
        if  z_array is None :
            warnings.warn('`Impedance` is None. could not compute'
                          ' `apparent resistivities`')
            raise CSex.pyCSAMTError_processing('`Impedance`  is None.'
                'Could not compute `resistivities`.'
                ' Please provided at least an `impedance array`.')
    
        resistivities = np.abs(z_array) ** 2 / (mu0 * omega_reffreq )
        
    elif resistivities is not None and z_array is None :
        
        if phase is None  and reference_freq is None:
            warnings.warn('Could not compute `impedance` without phase values'
                          ' and reference frequency!')
            raise CSex.pyCSAMTError_processing(
                'None Type could not be computed. '
                'Please provided phase values |reference frequency')
            
        z_array = np.sqrt(resistivities * omega_reffreq * mu0 ) * (
            np.cos(phase)+  1j * np.sin( phase) ) 
    
    # compute skin depths base of uncorrected apparent resistivities
    skin_depths = 503. *  np.sqrt(resistivities/reference_freq)

    # window_width = number_of_skin_depth * skin_depth
    def adapted_skindepth (skin_window, scale_value=10.):
        """
        Adapted windo width and ans scale value to adapt value to be 
        close to dipole length 
        
        :param skin_window: array of skin depth 1D botstick penetration 
                or window width 
        :type skin_window: array_like 
        
        :return: window width scaled and adapted 
        :rtype: array_like 
        
        """
        # Adjusted each filter  iteratively at each station
        window_width = np.array([
            func.round_dipole_length(value, round_value=scale_value) # round dipole 
                                 for value in skin_window ])
        #adjusting filter width to with ten to ten range 
        for ii,  ww in enumerate(window_width): 
            if  ii ==0 or ii == len(window_width)-1:
                if ww < dipole_length:
                    window_width[ii]=func.round_dipole_length(
                        dipole_length, round_value=scale_value)
            elif ww < dipole_length:
                window_width[ii]=func.round_dipole_length(
                    window_width[ii-1:ii+2].mean(), round_value =scale_value)

        return  window_width
    
    def compute_filter_length(skindepth, ref_freq, 
                              dipole_length=dipole_length ):
        """
        :param skindepth: Bostick skin depth calculated with
                    apparent ressistivities at each station 
        :type skin_depth: array_like, 
        
        :param ref_freq: frequency at clean data at each station 
        :type ref_freq: float
        
        :param dipole_length: dipole length of the survey in meters 
        :type dipole_length: float, default is 50.
        
        :return: weight factor beta j at the station 
        :type: array_like 
        
        """
        npoints =np.around(skindepth/dipole_length) 
        # adjusted width  and recompute  and adapthed width 
        npoints_array = np.arange(0, dipole_length * npoints, dipole_length) 
                                    
        
        bj= np.array([compute_weight_factor_betaj(val, dipole_length=dipole_length, 
                                                  number_of_points=npoints)[0]
                      for val in  npoints_array])
  
        return bj, bj.sum()
    
    # collect list of each factor of all  station 
    skin_depths= adapted_skindepth(skin_window=skin_depths,
                                   scale_value=adaptative_value)

    weight_factors= [compute_filter_length(skindepth = skin, ref_freq=reference_freq, 
                              dipole_length=dipole_length )[0] for skin in skin_depths]
    
    # compute new impedance with base impedance and  list of weight factors 
    zxy_impedancef = compute_zxy_xk_omega(z_imp=z_array, coeffs_bj= weight_factors)
    
    # recompute new apparent resistivity 
    resistivities =np.abs( zxy_impedancef) ** 2 / (mu0 * omega_reffreq )
    
    # now recompute new window width that and multiply with arbiraty constant c
    z_bostick_depth = np.abs(zxy_impedancef) / (omega_reffreq * mu0)
    
    z_bostick_depth= adapted_skindepth(skin_window=z_bostick_depth,
                                    scale_value=adaptative_value)
    
    # adaptative window width with number of skin depth 
    w_window_width = number_of_skin_depth * z_bostick_depth 
    
    # recompute again impedance so to get adaptative impedance 
    n_weight_factors= [compute_filter_length(skindepth = skin, ref_freq=reference_freq, 
                              dipole_length=dipole_length )[0] for skin in w_window_width]
    
    #-----------------------text will FLMA computation --------
    # npoints= int(w_window_width.mean()/dipole_length)
    # cAMA= compute_FLMA(z_array=zxy_impedancef, dipole_length=dipole_length, 
    #                     number_of_points= npoints)
    #------------------------------------------------------------
    cAMA = compute_zxy_xk_omega(z_imp=z_array, coeffs_bj= n_weight_factors)
    
    print('** {0:<27} {1} {2} m.'.format("Filter's width",
                                '=', np.around(w_window_width.mean(),2)))    

    return cAMA  

def compute_zxy_xk_omega(z_imp, coeffs_bj):
    """
    Compute adaptative impedance with variable length Hanning window
    :param z_imp: impedance value to be filtered 
    :type z_imp: complex 
    
    :param coeff_bj: list coefficients of weighted window at each stations
    :type coeffs_bj : list 
    
    :return: array of  new impedance filtered 
    :rtype: complex 
    
    :Example: 
        
        >>> from pycsamt.ff.processing import z_calculator as Zcc 
        >>> z1= np.array([2.46073791 +3.00162006j ,
        ...     9.74193019 +1.82209497j,
        ...     15.68879141 + 12.91164264j ,
        ...     5.84384925 +3.6899018j, 
        ...     2.4430065  +0.57175607j])
        >>> weight_betaj = [array([0.18169011, 0.81830989]), 
        ...                array([0.02492092, 0.25, 0.47507908, 0.25]),
        ...                array([0.00224272, 0.02771308, 0.09220631, 0.16554531, 0.21341394,
        ...                0.21341394, 0.16554531, 0.09220631, 0.02771308]),
        ...                 array([0.05766889, 0.47116556, 0.47116556]),
        ...                 array([1.])]
        >>> Zcc.compute_zxy_xk_omega(coeffs_bj=weight_betaj , z_imp=z1)
        ... [2.01364616+2.45625537j 9.16556955+4.84395488j 6.83942027+4.21606008j
        ...     4.80923612+2.75254645j 2.4430065 +0.57175607j]   
                      
    """
    znew_f=[]
    for kk, zz in enumerate(z_imp):
        wlen=len(coeffs_bj[kk]) 
        xk_cp=int(wlen/2)
        try :
            if wlen%2 !=0:
                ss=z_imp[kk- xk_cp:kk+ xk_cp+1] * coeffs_bj[kk] # for Python index 
            elif wlen%2 ==0 :
                 ss=z_imp[kk- xk_cp:kk+ xk_cp] * coeffs_bj[kk]
        except:
            # check back lengh and forelength so to fill outside window with wk=0 
            b0 = len(coeffs_bj[kk][:xk_cp]) - len(z_imp[:kk]) # control the pad backward
            f0 = len(coeffs_bj[kk][xk_cp:]) - len(z_imp[kk:]) # control the pad forward 
            if b0 >0:               # if pad doesnt not meet 
                bnew= np.concatenate((np.repeat(0., b0),z_imp[:kk])) # add 0 outside 
            else :bnew = z_imp[kk-xk_cp:kk]
            if f0 >0:
                fnew= np.concatenate((z_imp[kk:], np.repeat(0., f0)))#add 0 outside the window
            else :
                if wlen %2 ==0 : fnew=z_imp[kk: kk + xk_cp] 
                else : fnew=z_imp[kk: kk + xk_cp +1] 
    
            ss = np.concatenate((bnew, fnew))* coeffs_bj[kk]
    
        znew_f.append(ss.sum())
                
    return np.array(znew_f)

                          
if __name__=='__main__': 
    
    z1= np.array([2.46073791 +3.00162006j ,
         9.74193019 +1.82209497j,
         15.68879141 + 12.91164264j ,
         5.84384925 +3.6899018j, 
         2.4430065  +0.57175607j])
    
    reference_frequency = 8192.
    resistivities = np.abs(z1) ** 2 / (4* np.pi*1e-7 * 2* np.pi * reference_frequency )
    dipole_length =50
    znew= compute_AMA(reference_freq=reference_frequency,
                       number_of_skin_depth= 1., dipole_length =50.,
                       z_array=z1)
    print(resistivities)
    new_resistivities = np.abs(znew) ** 2 / (4* np.pi*1e-7 * 2* np.pi * reference_frequency )
    print(new_resistivities)
  
    


# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.

===============================================================================  

    .Script to UPDATE and UPGRADE GeoDataBase. GeoDataBase is the Core of geodrill module.
     More than 150 rocks as well as their properties have already set. User can
     add more informations to  increase the chance to plot a truth  pseudostratigraphy log
     with their right pattern , colors and resistivity values.  In addition , provide 
     the right resistivity value of  differents geological rocks,  may increase the accuracy
     to find the closest layer when only input resistivity values  are set (_see_module `Geodrill`)
    . pyCSAMT geoDataBase is flexible , can be edited at least you  are aware of what you are doing.
   
     GeoDatabse deals with all geological informations and  The electrical properties of rocks
     their pattern and the Digital Cartography map symbolization. 
     The database code colums 
          ::: ('code','label','__description','pattern', 'pat_size',	'pat_density',
            'pat_thickness','rgb', 'electrical_props', 'hatch', 'colorMPL', 'FGDC') ::: 
              
           ! `__description` can be replaced  by `name` !
     --Have fun ---
 
Created on Tue Feb 16 11:37:47 2021

@author: @Daniel03
"""

from pycsamt.geodrill.geoDB.sql_recorder import GeoDataBase 


# code and Label are automatically , dont need to add info . if provided , it will be rejected 
GET_INFO =False                                 # see the information already set in DataBase 
                                                # before nany update and upgrade , it is better to set this to true to see 
                                                # if the structure exist and what are the corresponding values in database .
UPDATE = False                                  # if update provide the geostructure name and the columns to modify 
UPGRADE =False                                  # if upgrade provide all  informations 

name_geo ='amethise '                           # name of stucture 
pattern_geo = 258.                              # In fact pattern are linked with hatch . the size of elements 
pat_size_geo = 1.5                              # size of hatch symbol (float) 
pat_density_geo =.23                            # density of hatch symbol (float)
pat_thickness = 2.                              # thick ness of hatch symbol (float)
rgb_geo ='R128G52B26'                           # RGB colors : eg RG28 or hexadecima #000ff  or 'blue'. program will convert to "RGB" except alpha
hatch_geo ='+.//.Oo'                            # matplotlib pattern build with symbolisation map ["/", "\\", "|", '-', '+', 'x', 'o', 'O', '.', '*'] 
electrical_props = [1e5 ,  1e3]                 # range of electrical property of rocks (maximum and minium)
colorMPL = (1.0, 0.25, 0.)                      # if rgb is provided , this key will set automatically , dont need to add 
FGDC ='none'                                    # Dont need to fill this part , Digital cartographic 
                                                # Standard for Geological  Map Symbolisation, will be fill later  at least you 
                                                # know what you do.
    

    
# fill the update dictionnary  and set UPDATE to TRUE  # update informationalready exist in database 
update_geo_dict = {'name': name_geo, 
                   'electrical_props':electrical_props}
# fill the upgrade dictionnary  and set UPGRADE to TRUE  # set new information to dataBase 
upgrade_geo_dict = {
                    'name': name_geo, 
                    'pattern': pattern_geo, 
                    'pat_size':pat_size_geo, 
                    'pat_density':pat_density_geo, 
                    'pat_thickness': pat_thickness,
                    'electrical_props':electrical_props , 
                    'hatch': hatch_geo , 
                    'colorMPL': colorMPL, 
                    }

# GET INFO ALREADY EXISTS 
if GET_INFO is True : 
    GeoDataBase()._reminder_geo_recorder(name_geo) # set the name you want to check  eg : GeoDataBase()._reminder_geo_recorder('amphibolite') 
#update INFO ALREADY EXISTS
if UPDATE is True : 
    rgeoDB= GeoDataBase()._update_geo_structure(**update_geo_dict)

#UPGRADE NEW STRUCTURE
if UPGRADE is True : 
    geodatabase_obj= GeoDataBase._add_geo_structure( **upgrade_geo_dict )
        

                



# -*- coding: utf-8 -*-
"""
Created on Mon Aug  3 14:36:23 2020

@author: @Daniel03

sets up input files for running 2d occam inversions adding elevation profile 

"""
import os  
import numpy as np 
import glob 


class Topo: 
    """
    Build numpy array  to add elevation profile and elevation model. 
    
    Attributes :
    ===========
        * add_elevation_profile : np:ndarray (2, num_elev_points)
            x=relative coordinates (offest ) , y= elevation   
        * add_elevation_model : np.nadarry (3,num_elev_points) 
            x=easting (UTM), northing(UTM) , elevation 
        * path_elev : is the path where the profiles files is located .
            ex: os.chdir(os.path.dirname())
                os.path.join(os.path.dirname(), K1_exp.bln)
    
        * step_betw_station : is offset between two stations .
        * self.prof_loc_east : Easting value (longitude(1,))
        * self.prof_loc_north : Nothing value (latitude (1,))
        * self.elev_value     : Elevation  value (Elevation(1,))
    
    Methods : 
    =========
    add_topo : return num_elev_points,  np.ndarray(2, num_elev_points) 
                row1 : relative coordinates 
                row2 : elevation points 
    add_topo_model : return elevation_model , np.ndarray(3, elevation_model)
                row1 : easting coordinates 
                row2 : northing coordinates 
                row3 : elevation points 
    """
    encodage='utf-8'
    
    def __init__(self, path_elev="K1_exp.bln", step_betw_station=50, ): 
        self.path_elev=path_elev 
        self.step_sta=step_betw_station
        
        self.prof_loc_east=None
        self.prof_loc_north=None
        self.elev_value=None
        
    
    def add_topo(self) : 
        """
        fonction  to add elevation from topography 
        x is the realtive offset 0  to n*step_betw_station 
        """
    
        file_elev=os.path.basename(self.path_elev)
        
        try: 
            
            ofi=open(file_elev,'r', encoding=self.encodage)
            
        except: 
            print("*** No such file in the directory***")
            return 
        exten_file=file_elev[-3:].upper()
        if file_elev[-3:] != exten_file: 
            print("** {} is a wrong format , should be *.bln or .BLN **".format(file_elev))

        head=ofi.readline() # read of headfile , not necessary.
        repere="east_DMS north_DMS UTM_zone   east_utm"\
            "      north_utm   azim_m  elev_m "
        if repere not in head:
            print(" * Please select the appropriate *.bln-file *")
            return
        
        prof_loc_list_east, prof_loc_list_north, elev_value_list=[],[],[]
        
        while 1 : 
            ligne=ofi.readline()
            ligne=ligne[:-1] # or ligne = ligne.strip()
            if ligne=='' or ligne == '\n': 
                break
            ligne=ligne[:-1] 
            list_line=ligne.split()
            prof_loc_list_east.append(float(list_line[6]))
            prof_loc_list_north.append(float(list_line[7]))
            elev_value_list.append(float(list_line[9]))
        ofi.close()
        
        # building numpy array (2, num_elev_points)
        rela_coord_list=[ii for ii in range(0,self.step_sta*len(elev_value_list),self.step_sta)] 
        
        prof_loc_east=np.asarray(prof_loc_list_east)
        prof_loc_north=np.asarray(prof_loc_list_north)
        elev_value=np.asarray(elev_value_list)
        rela_coord=np.asarray(rela_coord_list)
        
        rela_coord=rela_coord.reshape((1,rela_coord.shape[0]))
        
        self.prof_loc_east= prof_loc_east.reshape((1, prof_loc_east.shape[0]))

        self.prof_loc_north= prof_loc_north.reshape((1, prof_loc_north.shape[0]))

        self.elev_value=elev_value.reshape((1,elev_value.shape[0]))
        
        num_elev_points=np.concatenate((rela_coord,self.elev_value), axis=0)
        
        min_,max_=min(elev_value_list), max(elev_value_list)
        indmin, indmax=elev_value_list.index(min_),elev_value_list.index(max_)
        
        #Affichage 
        print("="*55)
        print("Line{0:^55}".format(file_elev[:-8]).upper())
        print("="*55)
        print("Nombre de stations: {:>5}".format(len(prof_loc_list_east)))
        print("Elevation mininale: station {0:>4} ={1:>6} {2:>2}".format("S"+str(indmin),min_,"m." ))
        print("Elevation maximale: station {0:>4} ={1:>6} {2:>2}".format("S"+str(indmin) ,max_,'m.'))
        print("{:^55}".format('------'))
        return num_elev_points
        
    def add_topo_model(self): 
        """
        generate a elevation model ndarray(3, num_elev_points)
        easting , northing , and elevation """
        
        list_models=[self.prof_loc_east, self.prof_loc_north, self.elev_value]
        elevationModel=list_models[0]
        for array in list_models[1:]: 
            elevationModel=np.concatenate((elevationModel,array), axis=0)
            
        return elevationModel
    

# TEST : 

if __name__=='__main__': 
    # path= r"C:/Users/kouao/OneDrive/paper2/process_paper2/"+\
    #     "data_stn/stnToDat_exportfiles/K1_cor.DAT"
    # dir_= r"C:/Users/kouao/OneDrive/paper2/process_paper2/"+\
    #     "data_stn/stnToDat_exportfiles"
    dir_=r'E:/OneDrive/Python/CodesExercices/'+\
        'ex_avgfiles/data/process_paper2/data_edifiles/K1_edi/K1_exp.BLN' # change the edipath
    
    os.chdir(r'E:/OneDrive/Python/CodesExercices/'+\
    'ex_avgfiles/data/process_paper2/data_edifiles/K1_edi')    
    
    top=Topo(path_elev=dir_, 
             step_betw_station=50)
    num_=top.add_topo()
    mod_=top.add_topo_model()

